@@ -33,10 +33,11 @@
 #include <cstring>
 
 #include "cluster_tree.hpp"
-#include "interaction.hpp"
+#include "assembly.hpp"
 #include "rk_matrix.hpp"
 #include "lapack_operations.hpp"
 #include "lapack_overloads.hpp"
+#include "blas_overloads.hpp"
 #include "full_matrix.hpp"
 #include "common/context.hpp"
 #include "common/my_assert.h"
@@ -47,7 +48,9 @@
 # ifndef __INTEL_COMPILER
 #  define isnan _isnan
 # endif
-#elif defined(__MINGW32__)
+#elif __GLIBC__ == 2 && __GLIBC_MINOR__ < 23
+// https://sourceware.org/bugzilla/show_bug.cgi?id=19439
+#elif __cplusplus >= 201103L || !defined(__GLIBC__)
 using std::isnan;
 #endif
 
@@ -55,45 +58,73 @@ using std::vector;
 using std::min;
 using std::max;
 
+namespace hmat {
 
-/** Convenience class to lighten the getRow() / getCol() calls.
+
+/* Convenience class to lighten the getRow() / getCol() / assemble calls
+   and use information from block_info_t to speed up things for sparse and null blocks.
 */
 template<typename T>
 class ClusterAssemblyFunction {
-  const AssemblyFunction<T>& f;
-  void* handle;
-  const ClusterData* rows;
-  const ClusterData* cols;
+  const Function<T>& f;
 
 public:
-  ClusterAssemblyFunction(const AssemblyFunction<T>& _f,
-                        const ClusterData* _rows, const ClusterData* _cols)
-    : f(_f), handle(NULL), rows(_rows), cols(_cols) {
-    f.prepareBlock(rows, cols, &handle);
+  const ClusterData* rows;
+  const ClusterData* cols;
+  hmat_block_info_t info;
+  const AllocationObserver & allocationObserver_;
+  ClusterAssemblyFunction(const Function<T>& _f,
+                          const ClusterData* _rows, const ClusterData* _cols,
+                          const AllocationObserver & allocationObserver)
+    : f(_f), rows(_rows), cols(_cols), allocationObserver_(allocationObserver) {
+    f.prepareBlock(rows, cols, &info, allocationObserver_);
+    assert((info.user_data == NULL) == (info.release_user_data == NULL));
   }
   ~ClusterAssemblyFunction() {
-    f.releaseBlock(handle);
+    f.releaseBlock(&info, allocationObserver_);
   }
   void getRow(int index, Vector<typename Types<T>::dp>& result) const {
-    f.getRow(rows, cols, index, handle, &result);
+    if (info.block_type != hmat_block_sparse || !info.is_null_row(&info, index))
+      f.getRow(rows, cols, index, info.user_data, &result);
   }
   void getCol(int index, Vector<typename Types<T>::dp>& result) const {
-    f.getCol(rows, cols, index, handle, &result);
+    if (info.block_type != hmat_block_sparse || !info.is_null_col(&info, index))
+      f.getCol(rows, cols, index, info.user_data, &result);
+  }
+  FullMatrix<typename Types<T>::dp>* assemble() const {
+    if (info.block_type != hmat_block_null)
+      return f.assemble(rows, cols, &info, allocationObserver_) ;
+    else
+      // TODO return
+      return FullMatrix<typename Types<T>::dp>::Zero(rows->size(), cols->size());
   }
 private:
-  ClusterAssemblyFunction(ClusterAssemblyFunction&o) {}; // No copy
+  ClusterAssemblyFunction(ClusterAssemblyFunction&o) {} // No copy
 };
 
-template<typename T> static double squaredNorm(const T x);
-template<> double squaredNorm(const C_t x) {
-  return norm(x);
-}
-template<> double squaredNorm(const Z_t x) {
-  return norm(x);
-}
 template<typename T> double squaredNorm(const T x) {
   return x * x;
 }
+// Specializations for complex values
+template<> double squaredNorm(const C_t x) {
+// std::norm seems deadfully slow on Intel 15
+#ifdef __INTEL_COMPILER
+  const float x_r = x.real();
+  const float x_i = x.imag();
+  return x_r*x_r + x_i*x_i;
+#else
+  return std::norm(x);
+#endif
+}
+template<> double squaredNorm(const Z_t x) {
+#ifdef __INTEL_COMPILER
+  const double x_r = x.real();
+  const double x_i = x.imag();
+  return x_r*x_r + x_i*x_i;
+#else
+  return std::norm(x);
+#endif
+}
 
 template<typename T> static bool isZero(const Vector<T>& v) {
   int index = v.absoluteMaxIndex();
@@ -132,7 +163,7 @@ template<typename T> static void findMax(FullMatrix<T>* m, int& i, int& j) {
   const int rows = m->rows;
   for (int col = 0; col < cols; col++) {
     for (int row = 0; row < rows; row++) {
-      double norm = squaredNorm<T>(m->get(row, col));
+      const double norm = squaredNorm<T>(m->get(row, col));
       if (norm > maxNorm) {
         i = row;
         j = col;
@@ -152,7 +183,7 @@ template<typename T> static void findMax(FullMatrix<T>* m, int& i, int& j) {
   \return the index of the chosen column, or -1 if no column can be found.
  */
 template<typename T>
-static int findCol(ClusterAssemblyFunction<T>& block, vector<bool>& colFree,
+static int findCol(const ClusterAssemblyFunction<T>& block, vector<bool>& colFree,
                    Vector<typename Types<T>::dp>& col) {
   int colCount = colFree.size();
   bool found = false;
@@ -173,7 +204,7 @@ static int findCol(ClusterAssemblyFunction<T>& block, vector<bool>& colFree,
 
 
 template<typename T>
-static int findMinRow(ClusterAssemblyFunction<T>& block,
+static int findMinRow(const ClusterAssemblyFunction<T>& block,
                       vector<bool>& rowFree,
                       const vector<Vector<typename Types<T>::dp>*>& aCols,
                       const vector<Vector<typename Types<T>::dp>*>& bCols,
@@ -210,7 +241,7 @@ static int findMinRow(ClusterAssemblyFunction<T>& block,
 }
 
 template<typename T>
-static int findMinCol(ClusterAssemblyFunction<T>& block,
+static int findMinCol(const ClusterAssemblyFunction<T>& block,
                       vector<bool>& colFree,
                       const vector<Vector<typename Types<T>::dp>*>& aCols,
                       const vector<Vector<typename Types<T>::dp>*>& bCols,
@@ -247,16 +278,17 @@ static int findMinCol(ClusterAssemblyFunction<T>& block,
 
 
 template<typename T>
-RkMatrix<T>* compressMatrix(FullMatrix<T>* m, const ClusterData* rows,
-                            const ClusterData* cols) {
+RkMatrix<T>* compressMatrix(FullMatrix<T>* m, const IndexSet* rows,
+                            const IndexSet* cols) {
   DECLARE_CONTEXT;
-  myAssert((size_t) m->rows == rows->n);
-  myAssert((size_t) m->cols == cols->n);
-  myAssert(m->lda >= m->rows);
+  assert(m->rows == rows->size());
+  assert(m->cols == cols->size());
+  assert(m->lda >= m->rows);
 
+  //TODO replace with a case with m==NULL
   bool zeroMatrix = true;
-  for (size_t col = 0; col < m->cols; col++) {
-    Vector<T> v(m->m + m->rows * col, m->rows);
+  for (int col = 0; col < m->cols; col++) {
+    Vector<T> v(m->m + ((size_t) m->rows) * col, m->rows);
     zeroMatrix = zeroMatrix && isZero(v);
     if (!zeroMatrix) break;
   }
@@ -272,12 +304,15 @@ RkMatrix<T>* compressMatrix(FullMatrix<T>* m, const ClusterData* rows,
   Vector<double>* sigma = NULL;
 
   int info = truncatedSvd<T>(m, &u, &sigma, &vt);
-  strongAssert(info == 0);
+  HMAT_ASSERT(info == 0);
   // Control of the approximation
 
   int maxK = min(rowCount, colCount);
   int k = RkMatrix<T>::approx.findK(sigma->v, maxK, RkMatrix<T>::approx.assemblyEpsilon);
 
+  if(k == 0)
+    return new RkMatrix<T>(NULL, rows, NULL, cols, NoCompression);
+
   for (int col = 0; col < k; col++) {
     for (int row = 0; row < rowCount; row++) {
       u->get(row, col) *= sigma->v[col];
@@ -299,12 +334,12 @@ RkMatrix<T>* compressMatrix(FullMatrix<T>* m, const ClusterData* rows,
 
 template<typename T>
 static RkMatrix<typename Types<T>::dp>*
-compressSvd(const AssemblyFunction<T>& f,
-            const ClusterData* rows, const ClusterData* cols) {
+compressSvd(const ClusterAssemblyFunction<T>& block) {
   DECLARE_CONTEXT;
   typedef typename Types<T>::dp dp_t;
-  FullMatrix<dp_t>* m = f.assemble(rows, cols);
-  RkMatrix<dp_t>* result = compressMatrix(m, rows, cols);
+  // TODO: use ClusterAssemblyFunction to optimize with blockinfo_t
+  FullMatrix<dp_t>* m = block.assemble();
+  RkMatrix<dp_t>* result = compressMatrix(m, block.rows, block.cols);
   delete m;
   return result;
 }
@@ -312,12 +347,13 @@ compressSvd(const AssemblyFunction<T>& f,
 
 template<typename T>
 static RkMatrix<typename Types<T>::dp>*
-compressAcaFull(const AssemblyFunction<T>& f,
-                const ClusterData* rows, const ClusterData* cols) {
+compressAcaFull(const ClusterAssemblyFunction<T>& block) {
   DECLARE_CONTEXT;
   typedef typename Types<T>::dp dp_t;
-  FullMatrix<dp_t>* m = f.assemble(rows, cols);
+  // TODO: use ClusterAssemblyFunction to optimize with blockinfo_t
+  FullMatrix<dp_t>* m = block.assemble();
 
+  const double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
   double estimateSquaredNorm = 0;
   int maxK = min(m->rows, m->cols);
   if (RkMatrix<dp_t>::approx.k > 0) {
@@ -333,7 +369,7 @@ compressAcaFull(const AssemblyFunction<T>& f,
   for (nu = 0; nu < maxK; nu++) {
     int i_nu, j_nu;
     findMax(m, i_nu, j_nu);
-    dp_t delta = m->get(i_nu, j_nu);
+    const dp_t delta = m->get(i_nu, j_nu);
     if (squaredNorm(delta) == 0.) {
       break;
     }
@@ -345,9 +381,7 @@ compressAcaFull(const AssemblyFunction<T>& f,
       tmpB.get(j, nu) = m->get(i_nu, j) / delta;
     }
 
-    FullMatrix<dp_t> a_nu(tmpA.m + nu * tmpA.rows, tmpA.rows, 1);
-    FullMatrix<dp_t> b_nu(tmpB.m + nu * tmpB.rows, tmpB.rows, 1);
-    m->gemm('N', 'T', Constants<dp_t>::mone, &a_nu, &b_nu, Constants<dp_t>::pone);
+    proxy_cblas::ger(m->rows, m->cols, Constants<dp_t>::mone, tmpA.m + nu * tmpA.rows, 1, tmpB.m + nu * tmpB.rows, 1, m->m, m->rows);
 
     // Update the estimate norm
     // Let S_{k-1} be the previous estimate. We have (for the Frobenius norm):
@@ -366,15 +400,14 @@ compressAcaFull(const AssemblyFunction<T>& f,
         newEstimate += hmat::real(Vector<dp_t>::dot(&va_nu, &a_l) * Vector<dp_t>::dot(&vb_nu, &b_l));
       }
       estimateSquaredNorm += 2.0 * newEstimate;
-      double a_nu_norm = va_nu.norm();
-      double b_nu_norm = vb_nu.norm();
-      double ab_norm_2 = (a_nu_norm * a_nu_norm) * (b_nu_norm * b_nu_norm);
+      const double a_nu_norm_2 = va_nu.normSqr();
+      const double b_nu_norm_2 = vb_nu.normSqr();
+      const double ab_norm_2 = a_nu_norm_2 * b_nu_norm_2;
       estimateSquaredNorm += ab_norm_2;
 
       // Evaluate the stopping criterion
       // ||a_nu|| ||b_nu|| < epsilon * ||S_nu||
       // <=> ||a_nu||^2 ||b_nu||^2 < epsilon^2 ||S_nu||^2
-      double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
       if (ab_norm_2 < epsilon * epsilon * estimateSquaredNorm) {
         break;
       }
@@ -383,7 +416,7 @@ compressAcaFull(const AssemblyFunction<T>& f,
   delete m;
 
   if (nu == 0) {
-    return new RkMatrix<dp_t>(NULL, rows, NULL, cols, AcaFull);
+    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols, AcaFull);
   }
 
   FullMatrix<dp_t>* newA = new FullMatrix<dp_t>(tmpA.rows, nu);
@@ -393,20 +426,20 @@ compressAcaFull(const AssemblyFunction<T>& f,
   newB->clear();
   memcpy(newB->m, tmpB.m, sizeof(dp_t) * tmpB.rows * nu);
 
-  return new RkMatrix<dp_t>(newA, rows, newB, cols, AcaFull);
+  return new RkMatrix<dp_t>(newA, block.rows, newB, block.cols, AcaFull);
 }
 
 
 template<typename T>
 static RkMatrix<typename Types<T>::dp>*
-compressAcaPartial(const AssemblyFunction<T>& f,
-                   const ClusterData* rows, const ClusterData* cols) {
+compressAcaPartial(const ClusterAssemblyFunction<T>& block) {
   typedef typename Types<T>::dp dp_t;
-  double estimateSquaredNorm = 0;
-  ClusterAssemblyFunction<T> block(f, rows, cols);
 
-  const int rowCount = rows->n;
-  const int colCount = cols->n;
+  const double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
+  double estimateSquaredNorm = 0;
+
+  const int rowCount = block.rows->size();
+  const int colCount = block.cols->size();
   int maxK = min(rowCount, colCount);
   // Contains false for the rows that were already used as pivot
   vector<bool> rowFree(rowCount, true);
@@ -421,7 +454,7 @@ compressAcaPartial(const AssemblyFunction<T>& f,
   int k = 0;
 
   do {
-    Vector<dp_t>* bCol = new Vector<dp_t>(cols->n);
+    Vector<dp_t>* bCol = new Vector<dp_t>(block.cols->size());
     // Calculation of row I and its residue
     block.getRow(I, *bCol);
     updateRow(*bCol, I, bCols, aCols, k);
@@ -431,7 +464,7 @@ compressAcaPartial(const AssemblyFunction<T>& f,
     // Find max and argmax of the residue
     double maxNorm2 = 0.;
     for (int j = 0; j < colCount; j++) {
-      double norm2 = squaredNorm<dp_t>(bCol->v[j]);
+      const double norm2 = squaredNorm<dp_t>(bCol->v[j]);
       if (colFree[j] && norm2 > maxNorm2) {
         maxNorm2 = norm2;
         J = j;
@@ -452,7 +485,7 @@ compressAcaPartial(const AssemblyFunction<T>& f,
       bCols.push_back(bCol);
 
       // Compute column J and residue
-      Vector<dp_t>* aCol = new Vector<dp_t>(rows->n);
+      Vector<dp_t>* aCol = new Vector<dp_t>(block.rows->size());
       block.getCol(J, *aCol);
       updateCol(*aCol, J, aCols, bCols, k);
       colFree[J] = false;
@@ -461,7 +494,7 @@ compressAcaPartial(const AssemblyFunction<T>& f,
       // Find max and argmax of the residue
       maxNorm2 = 0.;
       for (int i = 0; i < rowCount; i++) {
-        double norm2 = squaredNorm<dp_t>(aCol->v[i]);
+        const double norm2 = squaredNorm<dp_t>(aCol->v[i]);
         if (rowFree[i] && norm2 > maxNorm2) {
           maxNorm2 = norm2;
           I = i;
@@ -477,16 +510,15 @@ compressAcaPartial(const AssemblyFunction<T>& f,
         newEstimate += hmat::real(Vector<dp_t>::dot(aCol, aCols[l]) * Vector<dp_t>::dot(bCol, bCols[l]));
       }
       estimateSquaredNorm += 2.0 * newEstimate;
-      double aColNorm = aCol->norm();
-      double bColNorm = bCol->norm();
-      double ab_norm_2 = (aColNorm * aColNorm) * (bColNorm * bColNorm);
+      const double aColNorm_2 = aCol->normSqr();
+      const double bColNorm_2 = bCol->normSqr();
+      const double ab_norm_2 = aColNorm_2 * bColNorm_2;
       estimateSquaredNorm += ab_norm_2;
       k++;
 
       // Evaluate the stopping criterion
       // ||a_nu|| ||b_nu|| < epsilon * ||S_nu||
       // <=> ||a_nu||^2 ||b_nu||^2 < epsilon^2 ||S_nu||^2
-      double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
       if (ab_norm_2 < epsilon * epsilon * estimateSquaredNorm) {
         break;
       }
@@ -495,13 +527,13 @@ compressAcaPartial(const AssemblyFunction<T>& f,
 
   FullMatrix<dp_t> *newA, *newB;
   if (k != 0) {
-    newA = new FullMatrix<dp_t>(rows->n, k);
+    newA = new FullMatrix<dp_t>(block.rows->size(), k);
     for (int i = 0; i < k; i++) {
       memcpy(newA->m + (i * newA->rows), aCols[i]->v, sizeof(dp_t) * newA->rows);
       delete aCols[i];
       aCols[i] = NULL;
     }
-    newB = new FullMatrix<dp_t>(cols->n, k);
+    newB = new FullMatrix<dp_t>(block.cols->size(), k);
     for (int i = 0; i < k; i++) {
       memcpy(newB->m + (i * newB->rows), bCols[i]->v, sizeof(dp_t) * newB->rows);
       delete bCols[i];
@@ -509,31 +541,30 @@ compressAcaPartial(const AssemblyFunction<T>& f,
     }
   } else {
     // If k == 0, block is only made of zeros.
-    return new RkMatrix<dp_t>(NULL, rows, NULL, cols, AcaPartial);
+    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols, AcaPartial);
   }
 
-  return new RkMatrix<dp_t>(newA, rows, newB, cols, AcaPartial);
+  return new RkMatrix<dp_t>(newA, block.rows, newB, block.cols, AcaPartial);
 }
 
 
 template<typename T>
 static RkMatrix<typename Types<T>::dp>*
-compressAcaPlus(const AssemblyFunction<T>& f,
-                const ClusterData* rows, const ClusterData* cols) {
+compressAcaPlus(const ClusterAssemblyFunction<T>& block) {
   typedef typename Types<T>::dp dp_t;
+  const double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
   double estimateSquaredNorm = 0;
   int i_ref, j_ref;
-  int rowCount = rows->n, colCount = cols->n;
+  int rowCount = block.rows->size(), colCount = block.cols->size();
   int maxK = min(rowCount, colCount);
   Vector<dp_t> bRef(colCount), aRef(rowCount);
   vector<bool> rowFree(rowCount, true), colFree(colCount, true);
   vector<Vector<dp_t>*> aCols, bCols;
-  ClusterAssemblyFunction<T> block(f, rows, cols);
 
   j_ref = findCol(block, colFree, aRef);
   if (j_ref == -1) {
 	// The block is completely zero.
-    return new RkMatrix<dp_t>(NULL, rows, NULL, cols, AcaPlus);
+    return new RkMatrix<dp_t>(NULL, block.rows, NULL, block.cols, AcaPlus);
   }
 
   // The reference row is chosen such that it intersects the reference
@@ -560,7 +591,7 @@ compressAcaPlus(const AssemblyFunction<T>& f,
       updateRow<dp_t>(*bVec, i_star, bCols, aCols, k);
       j_star = bVec->absoluteMaxIndex();
       dp_t pivot = bVec->v[j_star];
-      strongAssert(pivot != Constants<dp_t>::zero);
+      HMAT_ASSERT(pivot != Constants<dp_t>::zero);
       // Calculate a
       block.getCol(j_star, *aVec);
       updateCol<dp_t>(*aVec, j_star, aCols, bCols, k);
@@ -571,7 +602,7 @@ compressAcaPlus(const AssemblyFunction<T>& f,
       updateCol<dp_t>(*aVec, j_star, aCols, bCols, k);
       i_star = aVec->absoluteMaxIndex();
       dp_t pivot = aVec->v[i_star];
-      strongAssert(pivot != Constants<dp_t>::zero);
+      HMAT_ASSERT(pivot != Constants<dp_t>::zero);
       // Calculate b
       block.getRow(i_star, *bVec);
       updateRow<dp_t>(*bVec, i_star, bCols, aCols, k);
@@ -592,16 +623,15 @@ compressAcaPlus(const AssemblyFunction<T>& f,
       newEstimate += hmat::real(Vector<dp_t>::dot(aVec, aCols[l]) * Vector<dp_t>::dot(bVec, bCols[l]));
     }
     estimateSquaredNorm += 2.0 * newEstimate;
-    double aVecNorm = aVec->norm();
-    double bVecNorm = bVec->norm();
-    double ab_norm_2 = (aVecNorm * aVecNorm) * (bVecNorm * bVecNorm);
+    const double aVecNorm_2 = aVec->normSqr();
+    const double bVecNorm_2 = bVec->normSqr();
+    const double ab_norm_2 = aVecNorm_2 * bVecNorm_2;
     estimateSquaredNorm += ab_norm_2;
     k++;
 
     // Evaluate the stopping criterion
     // ||a_nu|| ||b_nu|| < epsilon * ||S_nu||
     // <=> ||a_nu||^2 ||b_nu||^2 < epsilon^2 ||S_nu||^2
-    double epsilon = RkMatrix<dp_t>::approx.assemblyEpsilon;
     if (ab_norm_2 < epsilon * epsilon * estimateSquaredNorm) {
       break;
     }
@@ -609,8 +639,8 @@ compressAcaPlus(const AssemblyFunction<T>& f,
     // Update of a_ref and b_ref
     aRef.axpy(Constants<dp_t>::mone * bCols[k - 1]->v[j_ref], aCols[k - 1]);
     bRef.axpy(Constants<dp_t>::mone * aCols[k - 1]->v[i_ref], bCols[k - 1]);
-    bool needNewA = isZero(aRef) || (j_star == j_ref);
-    bool needNewB = isZero(bRef) || (i_star == i_ref);
+    const bool needNewA = isZero(aRef) || (j_star == j_ref);
+    const bool needNewB = isZero(bRef) || (i_star == i_ref);
 
     // If the row or the column of reference have been already chosen as pivot,
     // we can not keep it and then we take one or two others.
@@ -652,119 +682,108 @@ compressAcaPlus(const AssemblyFunction<T>& f,
     }
   } while (k < maxK);
 
-  myAssert(k > 0);
-  FullMatrix<dp_t>* newA = new FullMatrix<dp_t>(rows->n, k);
+  assert(k > 0);
+  FullMatrix<dp_t>* newA = new FullMatrix<dp_t>(block.rows->size(), k);
   for (int i = 0; i < k; i++) {
     memcpy(newA->m + (i * newA->rows), aCols[i]->v, sizeof(dp_t) * newA->rows);
     delete aCols[i];
     aCols[i] = NULL;
   }
-  FullMatrix<dp_t>* newB = new FullMatrix<dp_t>(cols->n, k);
+  FullMatrix<dp_t>* newB = new FullMatrix<dp_t>(block.cols->size(), k);
   for (int i = 0; i < k; i++) {
     memcpy(newB->m + (i * newB->rows), bCols[i]->v, sizeof(dp_t) * newB->rows);
     delete bCols[i];
     bCols[i] = NULL;
   }
-  return new RkMatrix<dp_t>(newA, rows, newB, cols, AcaPlus);
+  return new RkMatrix<dp_t>(newA, block.rows, newB, block.cols, AcaPlus);
 }
 
 #include <iostream>
+
+template<typename T>
+RkMatrix<typename Types<T>::dp>* compressWithoutValidation(CompressionMethod method,
+                                                           const ClusterAssemblyFunction<T>& block) {
+  typedef typename Types<T>::dp dp_t;
+  RkMatrix<dp_t>* rk = NULL;
+  switch (method) {
+  case Svd:
+    rk = compressSvd(block);
+    break;
+  case AcaFull:
+    rk = compressAcaFull(block);
+    break;
+  case AcaPartial:
+    rk = compressAcaPartial(block);
+    break;
+  case AcaPlus:
+    rk = compressAcaPlus(block);
+    break;
+  case NoCompression:
+    // Must not happen
+    HMAT_ASSERT(false);
+    break;
+  }
+
+  return rk;
+}
+
+
 /* Appele par HMatrix<T>::assemble() */
 template<typename T>
 RkMatrix<typename Types<T>::dp>* compress(CompressionMethod method,
-                                          const AssemblyFunction<T>& f,
+                                          const Function<T>& f,
                                           const ClusterData* rows,
-                                          const ClusterData* cols) {
+                                          const ClusterData* cols,
+                                          const AllocationObserver & ao) {
   typedef typename Types<T>::dp dp_t;
   RkMatrix<dp_t>* rk = NULL;
+  ClusterAssemblyFunction<T> block(f, rows, cols, ao);
 
-  // Always compress the smallest blocks using an SVD. Small blocks tend to have
-  // a bad compression ratio anyways, and the SVD is not very costly in this
-  // case.
-  // TODO: allow the setting of the size limit ?
-  if (max(rows->n, cols->n) < 100) {
-    rk = compressSvd(f, rows, cols);
-  } else {
-    switch (method) {
-    case Svd:
-      rk = compressSvd(f, rows, cols);
-      break;
-    case AcaFull:
-      rk = compressAcaFull(f, rows, cols);
-      break;
-    case AcaPartial:
-      rk = compressAcaPartial(f, rows, cols);
-      break;
-    case AcaPlus:
-      rk = compressAcaPlus(f, rows, cols);
-      break;
-    case NoCompression:
-      // Should not happen
-      break;
-    }
-  }
+  rk = compressWithoutValidation(method, block);
 
   if (HMatrix<T>::validateCompression) {
-    FullMatrix<dp_t>* full = f.assemble(rows, cols);
+    FullMatrix<dp_t>* full = block.assemble();
     if (rk->a) rk->a->checkNan();
     if (rk->b) rk->b->checkNan();
     FullMatrix<dp_t>* rkFull = rk->eval();
-    double approxNorm = rkFull->norm();
-    double fullNorm = full->norm();
+    const double approxNorm = rkFull->norm();
+    const double fullNorm = full->norm();
 
     // If I meet a NaN, I save & leave
     // TODO : improve this behaviour
     if (isnan(approxNorm)) {
       rkFull->toFile("Rk");
       full->toFile("Full");
-      strongAssert(false);
+      HMAT_ASSERT(false);
     }
 
     rkFull->axpy(Constants<T>::mone, full);
     double diffNorm = rkFull->norm();
     if (diffNorm > HMatrix<T>::validationErrorThreshold * fullNorm ) {
-      std::cout << "["<< rows->offset<<","<<rows->offset+rows->n -1 <<"]x["<< cols->offset<<","<<cols->offset+cols->n-1 <<"]"<< std::endl
+      std::cout << rows->description() << "x" << cols->description() << std::endl
            << std::scientific
            << "|M|  = " << fullNorm << std::endl
            << "|Rk| = " << approxNorm << std::endl
-           << "|M - Rk| / |Rk| = " << diffNorm / fullNorm << std::endl
-           << "Rank = " << rk->k << " / " << min(full->rows, full->cols) << std::endl << std::endl;
+           << "|M - Rk| / |M| = " << diffNorm / fullNorm << std::endl
+           << "Rank = " << rk->rank() << " / " << min(full->rows, full->cols) << std::endl << std::endl;
 
       if (HMatrix<T>::validationReRun) {
         // Call compression a 2nd time, for debugging with gdb the work of the compression algorithm...
         RkMatrix<dp_t>* rk_bis = NULL;
 
-        if (max(rows->n, cols->n) < 100) {
-          rk_bis = compressSvd(f, rows, cols);
-        } else {
-          switch (method) {
-            case Svd:
-              rk_bis = compressSvd(f, rows, cols);
-              break;
-            case AcaFull:
-              rk_bis = compressAcaFull(f, rows, cols);
-              break;
-            case AcaPartial:
-              rk_bis = compressAcaPartial(f, rows, cols);
-              break;
-            case AcaPlus:
-              rk_bis = compressAcaPlus(f, rows, cols);
-              break;
-            case NoCompression:
-              // Should not happen
-              break;
-          }
-        }
+        rk_bis = compressWithoutValidation(method, block);
         delete rk_bis ;
       }
 
       if (HMatrix<T>::validationDump) {
         std::string filename;
         std::ostringstream convert;   // stream used for the conversion
-        convert << "["<< rows->offset<<","<<rows->offset+rows->n -1 <<"]x["<< cols->offset<<","<<cols->offset+cols->n-1 <<"]" ;
+        convert << rows->description() << "x" << cols->description()  ;
 
         filename = "Rk_";
         filename += convert.str(); // set 'Result' to the contents of the stream
+        delete rkFull;
+        rkFull = rk->eval();
         rkFull->toFile(filename.c_str());
         filename = "Full_"+convert.str(); // set 'Result' to the contents of the stream
         full->toFile(filename.c_str());
@@ -778,12 +797,15 @@ RkMatrix<typename Types<T>::dp>* compress(CompressionMethod method,
 }
 
 // Declaration of the used templates
-template RkMatrix<S_t>* compressMatrix(FullMatrix<S_t>* m, const ClusterData* rows, const ClusterData* cols);
-template RkMatrix<D_t>* compressMatrix(FullMatrix<D_t>* m, const ClusterData* rows, const ClusterData* cols);
-template RkMatrix<C_t>* compressMatrix(FullMatrix<C_t>* m, const ClusterData* rows, const ClusterData* cols);
-template RkMatrix<Z_t>* compressMatrix(FullMatrix<Z_t>* m, const ClusterData* rows, const ClusterData* cols);
+template RkMatrix<S_t>* compressMatrix(FullMatrix<S_t>* m, const IndexSet* rows, const IndexSet* cols);
+template RkMatrix<D_t>* compressMatrix(FullMatrix<D_t>* m, const IndexSet* rows, const IndexSet* cols);
+template RkMatrix<C_t>* compressMatrix(FullMatrix<C_t>* m, const IndexSet* rows, const IndexSet* cols);
+template RkMatrix<Z_t>* compressMatrix(FullMatrix<Z_t>* m, const IndexSet* rows, const IndexSet* cols);
+
+template RkMatrix<Types<S_t>::dp>* compress<S_t>(CompressionMethod method, const Function<S_t>& f, const ClusterData* rows, const ClusterData* cols, const AllocationObserver &);
+template RkMatrix<Types<D_t>::dp>* compress<D_t>(CompressionMethod method, const Function<D_t>& f, const ClusterData* rows, const ClusterData* cols, const AllocationObserver &);
+template RkMatrix<Types<C_t>::dp>* compress<C_t>(CompressionMethod method, const Function<C_t>& f, const ClusterData* rows, const ClusterData* cols, const AllocationObserver &);
+template RkMatrix<Types<Z_t>::dp>* compress<Z_t>(CompressionMethod method, const Function<Z_t>& f, const ClusterData* rows, const ClusterData* cols, const AllocationObserver &);
+
+}  // end namespace hmat
 
-template RkMatrix<Types<S_t>::dp>* compress<S_t>(CompressionMethod method, const AssemblyFunction<S_t>& f, const ClusterData* rows, const ClusterData* cols);
-template RkMatrix<Types<D_t>::dp>* compress<D_t>(CompressionMethod method, const AssemblyFunction<D_t>& f, const ClusterData* rows, const ClusterData* cols);
-template RkMatrix<Types<C_t>::dp>* compress<C_t>(CompressionMethod method, const AssemblyFunction<C_t>& f, const ClusterData* rows, const ClusterData* cols);
-template RkMatrix<Types<Z_t>::dp>* compress<Z_t>(CompressionMethod method, const AssemblyFunction<Z_t>& f, const ClusterData* rows, const ClusterData* cols);
