@@ -62,25 +62,34 @@
 
 #include <stdlib.h>
 
+#ifdef HAVE_JEMALLOC
+#define JEMALLOC_NO_DEMANGLE
+#include <jemalloc/jemalloc.h>
+#endif
+
 #ifdef _MSC_VER
 // Intel compiler defines isnan in global namespace
 // MSVC defines _isnan
 # ifndef __INTEL_COMPILER
 #  define isnan _isnan
 # endif
-#elif defined(__MINGW32__)
+#elif __GLIBC__ == 2 && __GLIBC_MINOR__ < 23
+// https://sourceware.org/bugzilla/show_bug.cgi?id=19439
+#elif __cplusplus >= 201103L || !defined(__GLIBC__)
 using std::isnan;
 #endif
 
+namespace hmat {
+
 /** FullMatrix */
 template<typename T>
-FullMatrix<T>::FullMatrix(T* _m, size_t _rows, size_t _cols, int _lda)
-  : ownsMemory(false), m(_m), rows(_rows), cols(_cols), lda(_lda), pivots(NULL),
-    diagonal(NULL), isTriUpper(false), isTriLower(false) {
+FullMatrix<T>::FullMatrix(T* _m, int _rows, int _cols, int _lda)
+  : ownsMemory(false), triUpper_(false), triLower_(false),
+    m(_m), rows(_rows), cols(_cols), lda(_lda), pivots(NULL), diagonal(NULL) {
   if (lda == -1) {
     lda = rows;
   }
-  myAssert(lda >= rows);
+  assert(lda >= rows);
 }
 
 // #define POISON_ALLOCATION
@@ -114,13 +123,17 @@ template<> static void poisonArray(Z_t* array, size_t n) {
 #endif
 
 template<typename T>
-FullMatrix<T>::FullMatrix(size_t _rows, size_t _cols)
-  : ownsMemory(true), rows(_rows), cols(_cols), lda(_rows), pivots(NULL),
-    diagonal(NULL), isTriUpper(false), isTriLower(false) {
+FullMatrix<T>::FullMatrix(int _rows, int _cols)
+  : ownsMemory(true), triUpper_(false), triLower_(false),
+    rows(_rows), cols(_cols), lda(_rows), pivots(NULL), diagonal(NULL) {
   size_t size = ((size_t) rows) * cols * sizeof(T);
+#ifdef HAVE_JEMALLOC
+  m = (T*) je_calloc(size, 1);
+#else
   m = (T*) calloc(size, 1);
-  REGISTER_ALLOC(m, size);
-  strongAssert(m);
+#endif
+  HMAT_ASSERT_MSG(m, "Trying to allocate %ldb of memory failed (rows=%d cols=%d sizeof(T)=%d)", size, rows, cols, sizeof(T));
+  MemoryInstrumenter::instance().alloc(size, MemoryInstrumenter::FULL_MATRIX);
 #ifdef POISON_ALLOCATION
   // This memory is not initialized, fill it with NaNs to force a
   // crash when using it.
@@ -129,7 +142,7 @@ FullMatrix<T>::FullMatrix(size_t _rows, size_t _cols)
 }
 
 template<typename T>
-FullMatrix<T>* FullMatrix<T>::Zero(size_t rows, size_t cols) {
+FullMatrix<T>* FullMatrix<T>::Zero(int rows, int cols) {
   FullMatrix<T>* result = new FullMatrix<T>(rows, cols);
 #ifdef POISON_ALLOCATION
   // The memory was poisoned in FullMatrix<T>::FullMatrix(), set it back to 0;
@@ -141,8 +154,12 @@ FullMatrix<T>* FullMatrix<T>::Zero(size_t rows, size_t cols) {
 template<typename T> FullMatrix<T>::~FullMatrix() {
   if (ownsMemory) {
     size_t size = ((size_t) rows) * cols * sizeof(T);
-    REGISTER_FREE(m, size);
+    MemoryInstrumenter::instance().free(size, MemoryInstrumenter::FULL_MATRIX);
+#ifdef HAVE_JEMALLOC
+    je_free(m);
+#else
     free(m);
+#endif
     m = NULL;
   }
   if (pivots) {
@@ -154,9 +171,24 @@ template<typename T> FullMatrix<T>::~FullMatrix() {
 }
 
 template<typename T> void FullMatrix<T>::clear() {
-  myAssert(lda == rows);
+  assert(lda == rows);
   size_t size = ((size_t) rows) * cols * sizeof(T);
   memset(m, 0, size);
+  if (diagonal) {
+    memset(diagonal->v, 0, rows * sizeof(T));
+  }
+}
+
+template<typename T> size_t FullMatrix<T>::storedZeros() {
+  size_t result = 0;
+  for (int col = 0; col < cols; col++) {
+    for (int row = 0; row < rows; row++) {
+      if (std::abs(get(row, col)) < 1e-16) {
+        result++;
+      }
+    }
+  }
+  return result;
 }
 
 template<typename T> void FullMatrix<T>::scale(T alpha) {
@@ -165,8 +197,13 @@ template<typename T> void FullMatrix<T>::scale(T alpha) {
     if (alpha == Constants<T>::zero) {
       this->clear();
     } else {
-      // TODO: be careful to overflow!
-      int nm = rows * cols;
+      // Warning: check for overflow
+      size_t nm = ((size_t) rows) * cols;
+      const size_t block_size_blas = 1 << 30;
+      while (nm > block_size_blas) {
+        proxy_cblas::scal(block_size_blas, alpha, m + nm - block_size_blas, 1);
+        nm -= block_size_blas;
+      }
       proxy_cblas::scal(nm, alpha, m, 1);
     }
   } else {
@@ -183,13 +220,16 @@ template<typename T> void FullMatrix<T>::scale(T alpha) {
       }
     }
   }
+  if (diagonal) {
+    proxy_cblas::scal(rows, alpha, diagonal->v, 1);
+  }
 }
 
 template<typename T> void FullMatrix<T>::transpose() {
-  myAssert(lda == rows);
-  myAssert(m);
+  assert(lda == rows);
+  assert(m);
 #ifdef HAVE_MKL_IMATCOPY
-  proxy_mkl::imatcopy((size_t) rows, (size_t) cols, m);
+  proxy_mkl::imatcopy(rows, cols, m);
   std::swap(rows, cols);
   lda = rows;
 #else
@@ -214,28 +254,37 @@ template<typename T> void FullMatrix<T>::transpose() {
     }
   }
 #endif
-  if (isTriUpper) {
-    isTriUpper = false;
-    isTriLower = true;
-  } else if (isTriLower) {
-    isTriLower = false;
-    isTriUpper = true;
+  if (triUpper_) {
+    triUpper_ = false;
+    triLower_ = true;
+  } else if (triLower_) {
+    triLower_ = false;
+    triUpper_ = true;
   }
 }
 
-template<typename T> FullMatrix<T>* FullMatrix<T>::copy() const {
-  FullMatrix<T>* result = new FullMatrix<T>(rows, cols);
+template<typename T> FullMatrix<T>* FullMatrix<T>::copy(FullMatrix<T>* result) const {
+  if(result == NULL)
+    result = new FullMatrix<T>(rows, cols);
 
   if (lda == rows) {
     size_t size = ((size_t) rows) * cols * sizeof(T);
     memcpy(result->m, m, size);
   } else {
-    for (size_t col = 0; col < cols; col++) {
-      size_t resultOffset = result->rows * col;
-      size_t offset = lda * col;
+    for (int col = 0; col < cols; col++) {
+      size_t resultOffset = ((size_t) result->rows) * col;
+      size_t offset = ((size_t) lda) * col;
       memcpy(result->m + resultOffset, m + offset, rows * sizeof(T));
     }
   }
+
+  if (diagonal) {
+    if(!result->diagonal)
+      result->diagonal = new Vector<T>(rows);
+    memcpy(result->diagonal->v, diagonal->v, rows * sizeof(T));
+  }
+  result->triLower_ = triLower_;
+  result->triUpper_ = triUpper_;
   return result;
 }
 
@@ -247,9 +296,8 @@ template<typename T> FullMatrix<T>* FullMatrix<T>::copyAndTranspose() const {
     proxy_mkl::omatcopy(rows, cols, m, result->m);
   } else {
 #endif
-  const int _rows = rows, _cols = cols;
-  for (int i = 0; i < _rows; i++) {
-    for (int j = 0; j < _cols; j++) {
+  for (int i = 0; i < rows; i++) {
+    for (int j = 0; j < cols; j++) {
       result->get(j, i) = get(i, j);
     }
   }
@@ -267,10 +315,10 @@ void FullMatrix<T>::gemm(char transA, char transB, T alpha,
   int m = (transA == 'N' ? a->rows : a->cols);
   int n = (transB == 'N' ? b->cols : b->rows);
   int k = (transA == 'N' ? a->cols : a->rows);
-  myAssert(a->lda >= (transA == 'N' ? m : k));
-  myAssert(b->lda >= (transB == 'N' ? k : n));
-  myAssert(rows == m);
-  myAssert(cols == n);
+  assert(a->lda >= (transA == 'N' ? m : k));
+  assert(b->lda >= (transB == 'N' ? k : n));
+  assert(rows == m);
+  assert(cols == n);
   {
     const size_t _m = m, _n = n, _k = k;
     const size_t adds = _m * _n * _k;
@@ -283,8 +331,8 @@ void FullMatrix<T>::gemm(char transA, char transB, T alpha,
 
 template<typename T>
 void FullMatrix<T>::multiplyWithDiagOrDiagInv(const Vector<T>* d, bool inverse, bool left) {
-  myAssert(left || (this->cols == d->rows));
-  myAssert(!left || (this->rows == d->rows));
+  assert(left || (this->cols == d->rows));
+  assert(!left || (this->rows == d->rows));
 
   T* diag = d->v;
   {
@@ -296,7 +344,7 @@ void FullMatrix<T>::multiplyWithDiagOrDiagInv(const Vector<T>* d, bool inverse,
       // In this case, copying is a good idea since it avoids repeated
       // computations of 1 / diag[i].
       diag = (T*) malloc(d->rows * sizeof(T));
-      strongAssert(diag);
+      HMAT_ASSERT(diag);
       memcpy(diag, d->v, d->rows * sizeof(T));
       for (int i = 0; i < d->rows; i++) {
         diag[i] = Constants<T>::pone / diag[i];
@@ -305,7 +353,7 @@ void FullMatrix<T>::multiplyWithDiagOrDiagInv(const Vector<T>* d, bool inverse,
     // TODO: Test with scale to see if it is better.
     for (int j = 0; j < cols; j++) {
       for (int i = 0; i < rows; i++) {
-        m[i + j * lda] *= diag[i];
+        m[i + j * ((size_t) lda)] *= diag[i];
       }
     }
     if (inverse) {
@@ -313,17 +361,21 @@ void FullMatrix<T>::multiplyWithDiagOrDiagInv(const Vector<T>* d, bool inverse,
     }
   } else {
     for (int j = 0; j < cols; j++) {
-      proxy_cblas::scal(rows, diag[j], m + j * lda, 1);
+      T d = inverse ? Constants<T>::pone / diag[j] : diag[j];
+      proxy_cblas::scal(rows, d, m + j * ((size_t) lda), 1);
     }
   }
 }
 
 template<typename T>
 void FullMatrix<T>::ldltDecomposition() {
+  // Void matrix
+  if (rows == 0 || cols == 0) return;
+
   int n = this->rows;
   diagonal = new Vector<T>(n);
-  strongAssert(diagonal);
-  myAssert(this->rows == this->cols); // We expect a square matrix
+  HMAT_ASSERT(diagonal);
+  assert(this->rows == this->cols); // We expect a square matrix
 
   // Standard LDLt factorization algorithm is:
   //  diag[j] = A(j,j) - sum_{k < j} L(j,k)^2 diag[k]
@@ -332,14 +384,15 @@ void FullMatrix<T>::ldltDecomposition() {
   // An auxiliary array is introduced in order to perform less multiplications,
   // see  algorithm 1 in http://icl.cs.utk.edu/projectsfiles/plasma/pubs/ICL-UT-11-03.pdf
   T* v = new T[n];
-  strongAssert(v);
+  HMAT_ASSERT(v);
   for (int j = 0; j < n; j++) {
     for (int i = 0; i < j; i++)
       v[i] = get(j,i) * get(i,i);
 
     v[j] = get(j,j);
     for (int i = 0; i < j; i++)
-      v[j] -= get(j,i) * v[i];
+      // Do not use the -= operator because it's buggy in the intel compiler
+      v[j] = v[j] - get(j,i) * v[i];
 
     get(j,j) = v[j];
     for (int i = 0; i < j; i++)
@@ -347,6 +400,7 @@ void FullMatrix<T>::ldltDecomposition() {
         get(k,j) -= get(k,i) * v[i];
 
     for (int k = j+1; k < n; k++) {
+      HMAT_ASSERT_MSG(v[j] != Constants<T>::zero, "Division by 0 in LDLt");
       get(k,j) /= v[j];
     }
   }
@@ -358,14 +412,17 @@ void FullMatrix<T>::ldltDecomposition() {
       get(i,j) = Constants<T>::zero;
   }
 
-  isTriLower = true;
-  myAssert(!isTriUpper);
+  triLower_ = true;
+  assert(!isTriUpper());
   delete[] v;
 }
 
 template<typename T> void FullMatrix<T>::lltDecomposition() {
+    // Void matrix
+    if (rows == 0 || cols == 0) return;
+
     // from http://www.netlib.org/lapack/lawnspdf/lawn41.pdf page 120
-    const size_t n2 = rows * rows;
+    const size_t n2 = ((size_t) rows) * rows;
     const size_t n3 = n2 * rows;
     const size_t muls = n3 / 6 + n2 / 2 + rows / 3;
     const size_t adds = n3 / 6 - rows / 6;
@@ -374,8 +431,8 @@ template<typename T> void FullMatrix<T>::lltDecomposition() {
     int info = proxy_lapack::potrf('L', this->rows, this->m, this->lda);
     if(info != 0)
         // throw a pointer to be compliant with the Status class
-        throw new hmat::LapackException("potrf", info);
-    isTriLower = true;
+        throw hmat::LapackException("potrf", info);
+    triLower_ = true;
     for (int j = 0; j < this->cols; j++) {
         for(int i = 0; i < j; i++) {
             get(i,j) = Constants<T>::zero;
@@ -385,19 +442,20 @@ template<typename T> void FullMatrix<T>::lltDecomposition() {
 
 template<typename T>
 void FullMatrix<T>::luDecomposition() {
-  int mm = rows;
-  int n = cols;
+  // Void matrix
+  if (rows == 0 || cols == 0) return;
+
   pivots = (int*) calloc(rows, sizeof(int));
-  strongAssert(pivots);
+  HMAT_ASSERT(pivots);
   int info;
   {
-    const size_t _m = mm, _n = n;
+    const size_t _m = rows, _n = cols;
     const size_t muls = _m * _n *_n / 2 - _n *_n*_n / 6 + _m * _n / 2 - _n*_n / 2 + 2 * _n / 3;
     const size_t adds = _m * _n *_n / 2 - _n *_n*_n / 6 + _m * _n / 2 + _n / 6;
     increment_flops(Multipliers<T>::add * adds + Multipliers<T>::mul * muls);
   }
-  info = proxy_lapack::getrf(mm, n, m, lda, pivots);
-  strongAssert(!info);
+  info = proxy_lapack::getrf(rows, cols, m, lda, pivots);
+  HMAT_ASSERT(!info);
 }
 
 // The following code is very close to that of ZGETRS in LAPACK.
@@ -409,17 +467,19 @@ void FullMatrix<T>::luDecomposition() {
 // solve LX = (P ^ -1 B), which is done by ZLASWP with
 // the permutation. we used it just like in ZGETRS.
 template<typename T>
-void FullMatrix<T>::solveLowerTriangular(FullMatrix<T>* x) const {
-  myAssert(pivots || diagonal);
+void FullMatrix<T>::solveLowerTriangularLeft(FullMatrix<T>* x, bool unitriangular) const {
+  // Void matrix
+  if (x->rows == 0 || x->cols == 0) return;
+
   {
     const size_t _m = rows, _n = x->cols;
     const size_t adds = _n * _m * (_m - 1) / 2;
     const size_t muls = _n * _m * (_m + 1) / 2;
     increment_flops(Multipliers<T>::add * adds + Multipliers<T>::mul * muls);
   }
-  if (!diagonal && pivots)
+  if (pivots)
     proxy_lapack::laswp(x->cols, x->m, x->lda, 1, rows, pivots, 1);
-  proxy_cblas::trsm('L', 'L', 'N', 'U', rows, x->cols, Constants<T>::pone, m, lda, x->m, x->lda);
+  proxy_cblas::trsm('L', 'L', 'N', unitriangular ? 'U' : 'N', rows, x->cols, Constants<T>::pone, m, lda, x->m, x->lda);
 }
 
 
@@ -429,41 +489,41 @@ void FullMatrix<T>::solveLowerTriangular(FullMatrix<T>* x) const {
 //  the matrix was factorized before.
 
 template<typename T>
-void FullMatrix<T>::solveUpperTriangular(FullMatrix<T>* x, bool loweredStored) const {
+void FullMatrix<T>::solveUpperTriangularRight(FullMatrix<T>* x, bool unitriangular, bool lowerStored) const {
+  // Void matrix
+  if (x->rows == 0 || x->cols == 0) return;
+
   {
     const size_t _m = rows, _n = x->cols;
     const size_t adds = _n * _m * (_m - 1) / 2;
     const size_t muls = _n * _m * (_m + 1) / 2;
     increment_flops(Multipliers<T>::add * adds + Multipliers<T>::mul * muls);
   }
-  proxy_cblas::trsm('R', loweredStored ? 'L' : 'U', loweredStored ? 'T' : 'N', 'N',
+  proxy_cblas::trsm('R', lowerStored ? 'L' : 'U', lowerStored ? 'T' : 'N', unitriangular ? 'U' : 'N',
     x->rows, x->cols, Constants<T>::pone, m, lda, x->m, x->lda);
 }
 
 template<typename T>
-void FullMatrix<T>::solveUpperTriangularLeft(FullMatrix<T>* x, bool lowerStored) const {
-  myAssert(pivots || diagonal);
+void FullMatrix<T>::solveUpperTriangularLeft(FullMatrix<T>* x, bool unitriangular, bool lowerStored) const {
+  // Void matrix
+  if (x->rows == 0 || x->cols == 0) return;
+
   {
     const size_t _m = rows, _n = x->cols;
     const size_t adds = _n * _m * (_n - 1) / 2;
     const size_t muls = _n * _m * (_n + 1) / 2;
     increment_flops(Multipliers<T>::add * adds + Multipliers<T>::mul * muls);
   }
-  if (pivots) {
-    myAssert(!diagonal);
-    proxy_cblas::trsm('L', 'U', 'N', 'N',
-                rows, x->cols, Constants<T>::pone, m, lda, x->m, x->lda);
-  } else {
-    myAssert(diagonal);
-    myAssert(lowerStored);
-    proxy_cblas::trsm('L', 'L', 'T', 'U',
-                rows, x->cols, Constants<T>::pone, m, lda, x->m, x->lda);
-  }
+  proxy_cblas::trsm('L', lowerStored ? 'L' : 'U', lowerStored ? 'T' : 'N', unitriangular ? 'U' : 'N',
+    x->rows, x->cols, Constants<T>::pone, m, lda, x->m, x->lda);
 }
 
 template<typename T>
 void FullMatrix<T>::solve(FullMatrix<T>* x) const {
-  myAssert(pivots);
+  // Void matrix
+  if (x->rows == 0 || x->cols == 0) return;
+
+  assert(pivots);
   int ierr = 0;
   {
     const size_t nrhs = x->cols;
@@ -473,7 +533,7 @@ void FullMatrix<T>::solve(FullMatrix<T>* x) const {
     increment_flops(Multipliers<T>::add * adds + Multipliers<T>::mul * muls);
   }
   ierr = proxy_lapack::getrs('N', rows, x->cols, m, lda, pivots, x->m, x->rows);
-  strongAssert(!ierr);
+  HMAT_ASSERT(!ierr);
 }
 
 
@@ -483,14 +543,12 @@ void FullMatrix<T>::inverse() {
   // The inversion is done in two steps with dgetrf for LU decomposition and
   // dgetri for inversion of triangular matrices
 
-  myAssert(rows == cols);
+  assert(rows == cols);
 
-  int mm = rows;
-  int n = cols;
   int *ipiv = new int[rows];
   int info;
   {
-    size_t vn = n, vm = n;
+    size_t vn = cols, vm = cols;
     // getrf
     size_t additions = (vm*vn*vn)/2 - (vn*vn*vn)/6 - (vm*vn)/2 + vn/6;
     size_t multiplications = (vm*vn*vn)/2 - (vn*vn*vn)/6 + (vm*vn)/2
@@ -501,19 +559,19 @@ void FullMatrix<T>::inverse() {
     multiplications = (2*vn*vn*vn)/3 + (vn*vn)/2 + (5*vn)/6;
     increment_flops(Multipliers<T>::add * additions + Multipliers<T>::mul * multiplications);
   }
-  info = proxy_lapack::getrf(mm, n, m, lda, ipiv);
-  strongAssert(!info);
+  info = proxy_lapack::getrf(rows, cols, m, lda, ipiv);
+  HMAT_ASSERT(!info);
   // We call it twice: the first time to know the optimal size of
   // temporary arrays, and the second time for real calculation.
   int workSize;
   T workSize_req;
-  info = proxy_lapack::getri(mm, m, lda, ipiv, &workSize_req, -1);
+  info = proxy_lapack::getri(rows, m, lda, ipiv, &workSize_req, -1);
   workSize = (int) hmat::real(workSize_req) + 1;
   T* work = new T[workSize];
-  strongAssert(work);
-  info = proxy_lapack::getri(mm, m, lda, ipiv, work, workSize);
+  HMAT_ASSERT(work);
+  info = proxy_lapack::getri(rows, m, lda, ipiv, work, workSize);
   delete[] work;
-  strongAssert(!info);
+  HMAT_ASSERT(!info);
   delete[] ipiv;
 }
 
@@ -521,8 +579,8 @@ void FullMatrix<T>::inverse() {
 template<typename T>
 void FullMatrix<T>::copyMatrixAtOffset(const FullMatrix<T>* a,
                                        int rowOffset, int colOffset) {
-  myAssert(rowOffset + a->rows <= rows);
-  myAssert(colOffset + a->cols <= cols);
+  assert(rowOffset + a->rows <= rows);
+  assert(colOffset + a->cols <= cols);
 
 
   // Use memcpy when copying the whole matrix. This avoids BLAS calls.
@@ -544,8 +602,8 @@ template<typename T>
 void FullMatrix<T>::copyMatrixAtOffset(const FullMatrix<T>* a,
                                        int rowOffset, int colOffset,
                                        int rowsToCopy, int colsToCopy) {
-  myAssert(rowOffset + rowsToCopy <= rows);
-  myAssert(colOffset + colsToCopy <= cols);
+  assert(rowOffset + rowsToCopy <= rows);
+  assert(colOffset + colsToCopy <= cols);
   for (int col = 0; col < colsToCopy; col++) {
     proxy_cblas::copy(rowsToCopy, a->m + col * a->lda, 1,
                 (m + rowOffset + ((colOffset + col) * lda)), 1);
@@ -554,9 +612,9 @@ void FullMatrix<T>::copyMatrixAtOffset(const FullMatrix<T>* a,
 
 template<typename T>
 void FullMatrix<T>::axpy(T alpha, const FullMatrix<T>* a) {
-  myAssert(rows == a->rows);
-  myAssert(cols == a->cols);
-  size_t size = rows * cols;
+  assert(rows == a->rows);
+  assert(cols == a->cols);
+  size_t size = ((size_t) rows) * cols;
 
   increment_flops(Multipliers<T>::add * size
 		  + (alpha == Constants<T>::pone ? 0 : Multipliers<T>::mul * size));
@@ -567,49 +625,73 @@ void FullMatrix<T>::axpy(T alpha, const FullMatrix<T>* a) {
   }
 
   for (int col = 0; col < cols; col++) {
-    proxy_cblas::axpy(rows, alpha, a->m + col * a->lda, 1, m + col * lda, 1);
+    proxy_cblas::axpy(rows, alpha, a->m + ((size_t) col) * a->lda, 1, m + ((size_t) col) * lda, 1);
   }
 }
 
 template<typename T>
-double FullMatrix<T>::norm() const {
-  size_t size;
-  size = rows * cols;
+double FullMatrix<T>::normSqr() const {
+  size_t size = ((size_t) rows) * cols;
   T result = Constants<T>::zero;
 
   // Fast path
   if ((size < 1000000000) && (lda == rows)) {
     result += proxy_cblas_convenience::dot_c(size, m, 1, m, 1);
-    return sqrt(hmat::real(result));
+    return hmat::real(result);
   }
   for (int col = 0; col < cols; col++) {
     result += proxy_cblas_convenience::dot_c(rows, m + col * lda, 1, m + col * lda, 1);
   }
-  return sqrt(hmat::real(result));
+  return hmat::real(result);
 }
 
+template<typename T> double FullMatrix<T>::norm() const {
+  return sqrt(normSqr());
+}
+
+template<typename T> void FullMatrix<T>::fromFile(const char * filename) {
+  FILE * f = fopen(filename, "rb");
+  int code;
+  int r = fread(&code, sizeof(int), 1, f);
+  HMAT_ASSERT(r == 1);
+  HMAT_ASSERT(code == Constants<T>::code);
+  r = fread(&rows, sizeof(int), 1, f);
+  lda = rows;
+  HMAT_ASSERT(r == 1);
+  r = fread(&cols, sizeof(int), 1, f);
+  HMAT_ASSERT(r == 1);
+  r = fseek(f, 2 * sizeof(int), SEEK_CUR);
+  HMAT_ASSERT(r == 0);
+  if(m)
+      free(m);
+  size_t size = ((size_t) rows) * cols * sizeof(T);
+  m = (T*) calloc(size, 1);
+  r = fread(m, size, 1, f);
+  fclose(f);
+  HMAT_ASSERT(r == 1);
+}
 
 template<typename T> void FullMatrix<T>::toFile(const char *filename) const {
   int ierr;
   int fd;
-  size_t size = rows * cols * sizeof(T) + 5 * sizeof(int);
+  size_t size = ((size_t) rows) * cols * sizeof(T) + 5 * sizeof(int);
 
-  strongAssert(lda == rows);
+  HMAT_ASSERT(lda == rows);
 
   fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, (mode_t)0600);
-  strongAssert(fd != -1);
+  HMAT_ASSERT(fd != -1);
   ierr = lseek(fd, size - 1, SEEK_SET);
-  strongAssert(ierr != -1);
+  HMAT_ASSERT(ierr != -1);
   ierr = write(fd, "", 1);
-  strongAssert(ierr == 1);
+  HMAT_ASSERT(ierr == 1);
 #ifndef _WIN32
   void* mmapedFile = mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
   ierr = (mmapedFile == MAP_FAILED) ? 1 : 0;
-  strongAssert(!ierr);
+  HMAT_ASSERT(!ierr);
   int *asIntArray = (int*) mmapedFile;
   asIntArray[0] = Constants<T>::code;
-  asIntArray[1] = (int) rows;
-  asIntArray[2] = (int) cols;
+  asIntArray[1] = rows;
+  asIntArray[2] = cols;
   asIntArray[3] = sizeof(T);
   asIntArray[4] = 0;
   asIntArray += 5;
@@ -618,19 +700,19 @@ template<typename T> void FullMatrix<T>::toFile(const char *filename) const {
   close(fd);
   munmap(mmapedFile, size);
 #else
-#ifdef __GNUC__
-#warning MMAP: TO BE REMOVED OR FIXED
-#else
-#pragma message("Warning: MMAP: TO BE REMOVED OR FIXED")
-#endif
+  HMAT_ASSERT_MSG(false, "mmap not available on this platform");
 #endif
 
 }
 
+template<typename T> size_t FullMatrix<T>::memorySize() const {
+   return ((size_t) rows) * cols * sizeof(T);
+}
+
 template<typename T> void checkNanReal(const FullMatrix<T>* m) {
   for (int col = 0; col < m->cols; col++) {
     for (int row = 0; row < m->rows; row++) {
-      strongAssert(!isnan(m->get(row, col)));
+      HMAT_ASSERT(!isnan(m->get(row, col)));
     }
   }
 }
@@ -638,8 +720,8 @@ template<typename T> void checkNanReal(const FullMatrix<T>* m) {
 template<typename T> void checkNanComplex(const FullMatrix<T>* m) {
   for (int col = 0; col < m->cols; col++) {
     for (int row = 0; row < m->rows; row++) {
-      strongAssert(!isnan(m->get(row, col).real()));
-      strongAssert(!isnan(m->get(row, col).imag()));
+      HMAT_ASSERT(!isnan(m->get(row, col).real()));
+      HMAT_ASSERT(!isnan(m->get(row, col).imag()));
     }
   }
 }
@@ -660,27 +742,27 @@ template<> void FullMatrix<Z_t>::checkNan() const {
 
 // MmapedFullMatrix
 template<typename T>
-MmapedFullMatrix<T>::MmapedFullMatrix(size_t rows, size_t cols, const char* filename)
+MmapedFullMatrix<T>::MmapedFullMatrix(int rows, int cols, const char* filename)
   : m(NULL, rows, cols), mmapedFile(NULL), fd(-1), size(0) {
 #ifdef _WIN32
-  strongAssert(false); // no mmap() on Windows
+  HMAT_ASSERT(false); // no mmap() on Windows
 #else
   int ierr;
 
-  size = rows * cols * sizeof(T) + 5 * sizeof(int);
+  size = ((size_t) rows) * cols * sizeof(T) + 5 * sizeof(int);
   fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, (mode_t)0600);
-  strongAssert(fd != -1);
+  HMAT_ASSERT(fd != -1);
   ierr = lseek(fd, size - 1, SEEK_SET);
-  strongAssert(ierr != -1);
+  HMAT_ASSERT(ierr != -1);
   ierr = write(fd, "", 1);
-  strongAssert(ierr == 1);
+  HMAT_ASSERT(ierr == 1);
   mmapedFile = mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
   ierr = (mmapedFile == MAP_FAILED) ? 1 : 0;
-  strongAssert(!ierr);
+  HMAT_ASSERT(!ierr);
   int *asIntArray = (int*) mmapedFile;
   asIntArray[0] = 0;
-  asIntArray[1] = (int) rows;
-  asIntArray[2] = (int) cols;
+  asIntArray[1] = rows;
+  asIntArray[2] = cols;
   asIntArray[3] = sizeof(T);
   asIntArray[4] = 0;
   asIntArray += 5;
@@ -702,24 +784,24 @@ MmapedFullMatrix<T>* MmapedFullMatrix<T>::fromFile(const char* filename) {
   MmapedFullMatrix<T>* result = new MmapedFullMatrix();
 
 #ifdef _WIN32
-  strongAssert(false); // no mmap() on Windows
+  HMAT_ASSERT(false); // no mmap() on Windows
 #else
   int ierr;
   result->fd = open(filename, O_RDONLY);
-  strongAssert(result->fd != -1);
+  HMAT_ASSERT(result->fd != -1);
   struct stat fileStat;
   ierr = fstat(result->fd, &fileStat);
-  strongAssert(!ierr);
+  HMAT_ASSERT(!ierr);
   size_t fileSize = fileStat.st_size;
 
   result->mmapedFile = mmap(0, fileSize, PROT_READ, MAP_SHARED, result->fd, 0);
   ierr = (result->mmapedFile == MAP_FAILED) ? 1 : 0;
-  strongAssert(!ierr);
+  HMAT_ASSERT(!ierr);
   int* header = (int*) result->mmapedFile;
   // Check the consistency of the file
-  strongAssert(header[0] == Constants<T>::code);
-  strongAssert(header[3] == sizeof(T));
-  strongAssert(header[1] * ((size_t) header[2]) * sizeof(T) + (5 * sizeof(int)) == fileSize);
+  HMAT_ASSERT(header[0] == Constants<T>::code);
+  HMAT_ASSERT(header[3] == sizeof(T));
+  HMAT_ASSERT(header[1] * ((size_t) header[2]) * sizeof(T) + (5 * sizeof(int)) == fileSize);
   result->m.lda = result->m.rows = header[1];
   result->m.cols = header[2];
   result->m.m = (T*) (header + 5);
@@ -729,27 +811,27 @@ MmapedFullMatrix<T>* MmapedFullMatrix<T>::fromFile(const char* filename) {
 
 
 /** Vector */
-template<typename T> Vector<T>::Vector(T* _v, size_t _rows)
+template<typename T> Vector<T>::Vector(T* _v, int _rows)
   : ownsMemory(false), v(_v), rows(_rows) {}
 
-template<typename T> Vector<T>::Vector(size_t _rows)
+template<typename T> Vector<T>::Vector(int _rows)
   : ownsMemory(true), rows(_rows) {
-  size_t size = ((size_t) rows) * sizeof(T);
+  size_t size = rows * sizeof(T);
   v = (T*) calloc(size, 1);
-  REGISTER_ALLOC(v, size);
-  strongAssert(v);
+  MemoryInstrumenter::instance().alloc(size, MemoryInstrumenter::FULL_MATRIX);
+  HMAT_ASSERT(v);
 }
 
 template<typename T> Vector<T>::~Vector() {
   if (ownsMemory) {
-    size_t size = ((size_t) rows) * sizeof(T);
+    size_t size = rows * sizeof(T);
     free(v);
-    REGISTER_FREE(v, size);
+    MemoryInstrumenter::instance().free(size, MemoryInstrumenter::FULL_MATRIX);
   }
   v = NULL;
 }
 
-template<typename T> Vector<T>* Vector<T>::Zero(size_t rows) {
+template<typename T> Vector<T>* Vector<T>::Zero(int rows) {
   Vector<T> *result = new Vector<T>(rows);
   return result;
 }
@@ -767,18 +849,18 @@ void Vector<T>::gemv(char trans, T alpha,
   increment_flops(ops);
 
   if (trans == 'N') {
-    myAssert(rows == (size_t) a->rows);
-    myAssert(x->rows == (size_t) a->cols);
+    assert(rows == a->rows);
+    assert(x->rows == a->cols);
   } else {
-    myAssert(rows == (size_t) a->cols);
-    myAssert(x->rows == (size_t) a->rows);
+    assert(rows == a->cols);
+    assert(x->rows == a->rows);
   }
   proxy_cblas::gemv(t, matRows, matCols, alpha, a->m, lda, x->v, 1, beta, v, 1);
 }
 
 template<typename T>
 void Vector<T>::axpy(T alpha, const Vector* x) {
-  myAssert(rows == x->rows);
+  assert(rows == x->rows);
   proxy_cblas::axpy(rows, alpha, x->v, 1, this->v, 1);
 }
 
@@ -789,24 +871,28 @@ int Vector<T>::absoluteMaxIndex() const {
 
 template<typename T>
 T Vector<T>::dot(const Vector<T>* x, const Vector<T>* y) {
-  myAssert(x->rows == y->rows);
+  assert(x->rows == y->rows);
   // TODO: Beware of large vectors (>2 billion elements) !
   return proxy_cblas_convenience::dot_c(x->rows, x->v, 1, y->v, 1);
 }
 
 template<typename T> void Vector<T>::addToMe(const Vector<T>* x) {
-  myAssert(rows == x->rows);
+  assert(rows == x->rows);
   axpy(Constants<T>::pone, x);
 }
 
 template<typename T> void Vector<T>::subToMe(const Vector<T>* x) {
-  myAssert(rows == x->rows);
+  assert(rows == x->rows);
   axpy(Constants<T>::mone, x);
 }
 
 template<typename T> double Vector<T>::norm() const {
+  return sqrt(normSqr());
+}
+
+template<typename T> double Vector<T>::normSqr() const {
   T result = dot(this, this);
-  return sqrt(hmat::real(result));
+  return hmat::real(result);
 }
 
 template<typename T> void Vector<T>::clear() {
@@ -815,7 +901,7 @@ template<typename T> void Vector<T>::clear() {
 
 template<typename T> void Vector<T>::scale(T alpha) {
   if (alpha == Constants<T>::zero) {
-    memset(v, 0, sizeof(T) * ((size_t) rows));
+    memset(v, 0, sizeof(T) * rows);
   } else {
     proxy_cblas::scal(rows, alpha, v, 1);
   }
@@ -836,3 +922,5 @@ template class Vector<S_t>;
 template class Vector<D_t>;
 template class Vector<C_t>;
 template class Vector<Z_t>;
+
+}  // end namespace hmat
