@@ -28,15 +28,19 @@
 #define _H_MATRIX_HPP
 
 #include "tree.hpp"
-#include "interaction.hpp"
+#include "assembly.hpp"
 #include "data_types.hpp"
 #include "full_matrix.hpp"
 #include "cluster_tree.hpp"
+#include "admissibility.hpp"
+#include "recursion.hpp"
 #include <cassert>
 #include <fstream>
 #include <iostream>
 
 
+namespace hmat {
+
 template<typename T> class Vector;
 template<typename T> class RkMatrix;
 
@@ -44,6 +48,21 @@ template<typename T> class RkMatrix;
  */
 enum SymmetryFlag {kNotSymmetric, kLowerSymmetric};
 
+/** Default rank value for blocks that dont have an actual computed rank
+   */
+enum DefaultRank {UNINITIALIZED_BLOCK = -3, NONLEAF_BLOCK = -2, FULL_BLOCK = -1};
+
+/** Settings global to a whole matrix */
+struct MatrixSettings {
+};
+
+/** Settings local to a matrix bloc */
+struct LocalSettings {
+    const MatrixSettings * global;
+    explicit LocalSettings(const MatrixSettings * s): global(s) {}
+    //TODO add epsilons
+};
+
 /** Degrees of freedom permutation of a vector required in HMatrix context.
 
      In order that the subsets of rows and columns are
@@ -67,53 +86,96 @@ template<typename T> void reorderVector(FullMatrix<T>* v, int* indices);
  */
 template<typename T> void restoreVectorOrder(FullMatrix<T>* v, int *indices);
 
+template<typename T> class HMatrix;
+/** Class to write user defined data when dumping matrix onto disk.
 
-/*! \brief Data held by an HMatrix tree node.
+    This class is used by dumpTreeToFile to write extra information into
+    json file; the returned string must thus be a valid json fragment text.
  */
-template<typename T> class HMatrixData {
+template<typename T> class HMatrixNodeDumper {
 public:
-  /// Rows of this HMatrix block
-  ClusterTree * rows;
-  /// Columns of this HMatrix block
-  ClusterTree * cols;
-  /// Compressed block, or NULL if the block is not a leaf or is full.
-  RkMatrix<T> *rk;
-  /// Full block, or NULL if the block is not a leaf or is compressed.
-  FullMatrix<T> *m;
+  HMatrixNodeDumper() {}
+  virtual std::string dumpExtraInfo(const HMatrix<T>& node, const std::string& prefix) const = 0;
+};
 
+template<typename T>
+class HMatrixVoidNodeDumper : public HMatrixNodeDumper<T> {
 public:
-  HMatrixData() : rows(NULL), cols(NULL), rk(NULL), m(NULL) {}
-  HMatrixData(ClusterTree* _rows, ClusterTree* _cols)
-  : rows(_rows), cols(_cols), rk(NULL), m(NULL) {}
-  ~HMatrixData();
-  /*! \brief Return true if the block is admissible.
-   */
-  bool isAdmissibleLeaf() const {
-    return (rows->isAdmissibleWith(cols));
-  }
+  virtual std::string dumpExtraInfo(const HMatrix<T>&, const std::string&) const { return ""; }
+};
+
+/** Class to truncate Rk matrices.
+ */
+template<typename T>
+class EpsilonTruncate : public TreeProcedure<HMatrix<T> > {
+private:
+  double epsilon_;
+public:
+  EpsilonTruncate(double epsilon) : epsilon_(epsilon) {}
+  void visit(HMatrix<T> * node, const Visit order) const;
 };
 
 /*! \brief The HMatrix class, representing a HMatrix.
 
-  It is a tree of arity arity(ClusterTree)^2, 4 in this case.
+  It is a tree of arity arity(ClusterTree)^2, 4 in most cases.
   An HMatrix is a tree-like structure that is:
     - a Leaf : in this case the node is either really a RkMatrix
       (compressed block), or a small dense block.
     - an internal node : in this case, it has 4 children that form a partition
       of the HMatrix Dofs, and the node doesn't carry data itself.
  */
-template<typename T> class HMatrix : public Tree<4> {
+template<typename T> class HMatrix : public Tree<HMatrix<T> >, public RecursionMatrix<T, HMatrix<T> > {
   friend class RkMatrix<T>;
+
+  /// Rows of this HMatrix block
+  const ClusterTree * rows_;
+  /// Columns of this HMatrix block
+  const ClusterTree * cols_;
+  union {
+   /// Compressed block, or NULL if the block is not a leaf or is full.
+   RkMatrix<T> * rk_;
+   /// Full block, or NULL if the block is not a leaf or is compressed.
+   FullMatrix<T> * full_;
+  };
+  /// rank_ of the block for Rk matrices, or: UNINITIALIZED_BLOCK=-3 for an uninitialized matrix, NONLEAF_BLOCK=-2 for non leaf, FULL_BLOCK=-1 for full a matrix
+  int rank_;
+  void uncompatibleGemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b);
+  void recursiveGemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b);
+  void leafGemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b);
+  HMatrix<T> * fullRkSubset(const IndexSet* subset, bool col) const;
+  /*! \brief Auxiliary function used by HMatrix::dumpTreeToFile().
+   */
+  void dumpSubTree(std::ofstream& f, int depth, const HMatrixNodeDumper<T>& nodeDumper) const;
+  /** Only used by internalCopy */
+  HMatrix(const MatrixSettings * settings);
 public:
   /*! \brief Create a HMatrix based on a row and column ClusterTree.
 
     \param _rows The row cluster tree
     \param _cols The column cluster tree
    */
-  HMatrix(ClusterTree* _rows, ClusterTree* _cols, SymmetryFlag symFlag = kNotSymmetric);
+  HMatrix(ClusterTree* _rows, ClusterTree* _cols, const MatrixSettings * settings,
+       SymmetryFlag symmetryFlag = kNotSymmetric,
+       AdmissibilityCondition * admissibilityCondition = &StandardAdmissibilityCondition::DEFAULT_ADMISSIBLITY);
+
+  /*! \brief Create a copy of this matrix for internal use only.
+   * Only copy this node, not the whole tree. The created matrix
+   * is an uninitialized leaf with same rows and cols as this.
+   */
+  HMatrix<T> * internalCopy(bool temporary = false, bool withChildren = false) const;
+  ~HMatrix();
+
+  /*! \brief HMatrix coarsening.
+
+     If all children are Rk leaves, then we try to merge them into a single Rk-leaf.
+     This is done if the memory of the resulting leaf is less than the sum of the initial
+     leaves. Note that this operation could be used hierarchically.
+     \param upper the symmetric of 'this', when building a non-sym matrix with a sym content
+   */
+  void coarsen(HMatrix<T>* upper = NULL) ;
   /*! \brief HMatrix assembly.
    */
-  void assemble(const AssemblyFunction<T>& f);
+  void assemble(Assembly<T>& f, const AllocationObserver & = AllocationObserver());
   /*! \brief HMatrix assembly.
 
     \param f the assembly function
@@ -121,7 +183,9 @@ public:
                  that upper=this (that is, the current block is on the diagonal)
     \param onlyLower if true, only assemble the lower part of the matrix, ie don't copy.
    */
-  void assembleSymmetric(const AssemblyFunction<T>& f, HMatrix<T>* upper=NULL, bool onlyLower=false);
+  void assembleSymmetric(Assembly<T>& f,
+     HMatrix<T>* upper=NULL, bool onlyLower=false,
+     const AllocationObserver & = AllocationObserver());
   /*! \brief Evaluate the HMatrix, ie converts it to a full matrix.
 
     This conversion does the reorderng of the unknowns such that the resulting
@@ -142,31 +206,26 @@ public:
     \param _rows Rows of the result matrix
     \param _cols Columns of the result matrix
    */
-  void evalPart(FullMatrix<T>* result, const ClusterData* _rows, const ClusterData* _cols) const;
-  /*! Compute the compression ratio of the HMatrix.
+  void evalPart(FullMatrix<T>* result, const IndexSet* _rows, const IndexSet* _cols) const;
+
+  void info(hmat_info_t &);
 
-    \return the pair (elements_stored, total_elements).
-   */
-  std::pair<size_t, size_t> compressionRatio() const;
   /** This *= alpha
 
       \param alpha scaling factor
    */
   void scale(T alpha);
-  /** Compute y <- alpha * op(A) * y + beta * y.
+  /** Compute y <- alpha * op(this) * x + beta * y.
 
       The arguments are similar to BLAS GEMV.
    */
   void gemv(char trans, T alpha, const Vector<T>* x, T beta, Vector<T>* y) const;
-  /** Compute y <- alpha * op(A) * y + beta * y.
+  /** Compute y <- alpha * op(this) * x + beta * y.
 
       The arguments are similar to BLAS GEMV.
-
-      This version is in principle similar to GEMM(), but if you take
-      enough drug, it is not the same thing at all.
    */
   void gemv(char trans, T alpha, const FullMatrix<T>* x, T beta, FullMatrix<T>* y) const;
-  /*! \brief this <- alpha * A * B + beta * C
+  /*! \brief this <- alpha * op(A) * op(B) + beta * this
 
     \param transA 'N' or 'T', as in BLAS
     \param transB 'N' or 'T', as in BLAS
@@ -175,9 +234,9 @@ public:
     \param b the matrix B
     \param beta beta
    */
-  void gemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b, T beta, int depth=0);
-  /*! \brief S <- S - M * D * M^T, where S is symmetric (Lower stored),
-      D diagonal, this = S
+  void gemm(char transA, char transB, T alpha, const HMatrix<T>* a, const HMatrix<T>*b, T beta);
+  /*! \brief this <- this - M * D * M^T, where 'this' is symmetric (Lower stored),
+      D diagonal
 
       \warning D has to be reduced in ldlt form with d->ldltDecomposition() before
 
@@ -185,6 +244,17 @@ public:
       \param d D : only the diagonal of this matrix is considered
    */
   void mdmtProduct(const HMatrix<T> * m, const HMatrix<T> * d);
+
+  /*! \brief this <- this - M * D * N^T with D diagonal
+
+      \warning D has to be reduced in ldlt form with d->ldltDecomposition() before
+
+      \param m M
+      \param d D : only the diagonal of this matrix is considered
+      \param n N
+   */
+  void mdntProduct(const HMatrix<T>* m, const HMatrix<T>* d, const HMatrix<T>* n);
+
   /** Create a matrix filled with 0s, with the same structure as H.
 
       \param h the model matrix,
@@ -197,7 +267,10 @@ public:
       \param cols the column ClusterTree.
       \return a 0 HMatrix.
    */
-  static HMatrix<T>* Zero(const ClusterTree* rows, const ClusterTree* cols);
+  static HMatrix<T>* Zero(const ClusterTree* rows, const ClusterTree* cols,
+                          const MatrixSettings * settings,
+                          AdmissibilityCondition * admissibilityCondition =
+                          &StandardAdmissibilityCondition::DEFAULT_ADMISSIBLITY);
   /*! \brief Create a Postscript file representing the HMatrix.
 
     The result .ps file shows the matrix structure and the compression ratio. In
@@ -207,7 +280,7 @@ public:
 
     \param filename output filename.
    */
-  void createPostcriptFile(const char* filename) const;
+  void createPostcriptFile(const std::string& filename) const;
   /*! \brief Dump some HMatrix metadata to a Python-readable file.
 
     This function create a file that is readable by Python's eval()
@@ -226,10 +299,13 @@ public:
 
     \param filename path to the output file.
    */
-  void dumpTreeToFile(const char* filename) const;
+  void dumpTreeToFile(const std::string& filename, const HMatrixNodeDumper<T>& nodeDumper) const;
+  /** Returns a copy of this (with all the structure and data)
+       */
+  HMatrix<T>* copy() const ;
   /** this <- o (copy)
 
-      \param o The HMatrix t copy
+      \param o The HMatrix to copy. 'this' must be allready created and have the right structure.
    */
   void copy(const HMatrix<T>* o);
   /** Copy the structure of an HMatrix without copying its content.
@@ -238,9 +314,14 @@ public:
       allocated) mirroring the structure of this.
    */
   HMatrix<T>* copyStructure() const;
+  /*! \brief Return square of the Frobenius norm of the matrix.
+   */
+  double normSqr() const;
   /*! \brief Return the Frobenius norm of the matrix.
    */
-  double norm() const;
+  double norm() const {
+    return sqrt(normSqr());
+  }
   /** Set a matrix to 0.
    */
   void clear();
@@ -251,10 +332,17 @@ public:
       computation (and will be freed at the end).
       \param depth The depth, used for pretty printing purposes
    */
-  void inverse(HMatrix<T>* tmp=NULL, int depth=0);
+  void inverse();
   /*! \brief Transpose the H-matrix in place
    */
   void transpose();
+
+  /**
+   * Swap non diagonal blocks and cluster trees.
+   * Only used internally.
+   */
+  void transposeNoRecurse();
+
   /*! \brief this <- o^t
 
     \param o
@@ -271,15 +359,7 @@ public:
    */
   void ldltDecomposition();
   void lltDecomposition();
-private:
-  /*! \brief Auxiliary function used by HMatrix::dumpTreeToFile().
-   */
-  void dumpSubTree(std::ofstream& f, int depth) const;
 
-public:
-  /*! \brief Build a "fake" HMatrix for internal use only
-   */
-  HMatrix();
   /** This <- This + alpha * b
 
       \param alpha
@@ -299,28 +379,28 @@ public:
       \param rows
       \param cols
    */
-  void axpy(T alpha, const FullMatrix<T>* b, const ClusterData* rows, const ClusterData* cols);
+  void axpy(T alpha, const FullMatrix<T>* b, const IndexSet* rows, const IndexSet* cols);
+  /** This <- This + alpha * Id
+
+      \param alpha
+   */
+  void addIdentity(T alpha);
   /*! Return true if this is a full block.
    */
   inline bool isFullMatrix() const {
-    return isLeaf() && data.m;
-  };
+    return rank_ == FULL_BLOCK && full_ != NULL;
+  }
   /* Return the full matrix corresponding to the current leaf
    */
   FullMatrix<T>* getFullMatrix() const {
     assert(isFullMatrix());
-    return data.m;
+    return full_;
   }
   /*! Return true if this is a compressed block.
    */
   inline bool isRkMatrix() const {
-    return isLeaf() && data.rk;
-  };
-  /*! Return true if this is not a leaf.
-   */
-  inline bool isHMatrix() const {
-    return !isLeaf();
-  };
+    return rank_ >= 0;
+  }
   /*! \brief Return F * H (F Full, H divided)
    */
   static FullMatrix<T>* multiplyFullH(char transM, char transH, const FullMatrix<T>* m, const HMatrix<T>* h);
@@ -343,48 +423,43 @@ public:
       Le resultat est alors une matrice pleine.
   */
   static FullMatrix<T>* multiplyFullMatrix(char transA, char transB, const HMatrix<T>* a, const HMatrix<T>* b);
-  /*! \brief B <- B*D : ou B = this et D est en argument
+  /*! \brief B <- B*D where B is this
 
-    \warning D doit avoir ete decomposee en LDL^T avant
+    \warning D must a have been decomposed by LDLt
     \param d matrice D
+    \param left run B <- D*B instead of B <- B*D
+    \param inverse run B <- B * D^-1
   */
-  void multiplyWithDiag(const HMatrix<T>* d, bool left = false);
-  /*! \brief B <- B*D ou B <- B*D^-1  (ou idem a gauche) : ou B = this et D est en argument
-
-     \param d matrice D
-     \param inverse true : B<-B*D^-1, false B<-B*D
-     \param left true : B<-D*B, false B<-B*D
-  */
-  void multiplyWithDiagOrDiagInv(const HMatrix<T>* d, bool inverse, bool left = false);
+  void multiplyWithDiag(const HMatrix<T>* d, bool left = false, bool inverse = false) const;
   /*! \brief Resolution du systeme L X = B, avec this = L, et X = B.
 
     \param b la matrice B en entree, et X en sortie.
    */
-  void solveLowerTriangular(HMatrix<T>* b) const;
+  void solveLowerTriangularLeft(HMatrix<T>* b, bool unitriangular) const;
   /*! \brief Resolution du systeme L x = x, avec this = L, et x = b vecteur.
 
     B est un vecteur a plusieurs colonnes, donc une FullMatrix.
 
     \param b Le vecteur b en entree, et x en sortie.
    */
-  void solveLowerTriangular(FullMatrix<T>* b) const;
+  void solveLowerTriangularLeft(FullMatrix<T>* b, bool unitriangular) const;
   /*! Resolution de X U = B, avec U = this, et X = B.
 
     \param b la matrice B en entree, X en sortie
    */
-  void solveUpperTriangular(HMatrix<T>* b, bool loweredStored = false) const;
+  void solveUpperTriangularRight(HMatrix<T>* b, bool unitriangular, bool lowerStored) const;
   /*! Resolution de U X = B, avec U = this, et X = B.
 
     \param b la matrice B en entree, X en sortie
    */
-  void solveUpperTriangularLeft(HMatrix<T>* b) const;
+  void solveUpperTriangularLeft(HMatrix<T>* b, bool unitriangular, bool lowerStored) const;
   /*! Resolution de x U = b, avec U = this, et x = b.
 
     \warning b est un vecteur ligne et non colonne.
 
     \param b Le vecteur b en entree, x en sortie.
    */
-  void solveUpperTriangular(FullMatrix<T>* b, bool loweredStored = false) const;
+  void solveUpperTriangularRight(FullMatrix<T>* b, bool unitriangular, bool lowerStored) const;
   /*! Resolution de U x = b, avec U = this, et x = b.
     U peut etre en fait L^T ou L est une matrice stockee inferieurement
     en precisant lowerStored = true
@@ -393,12 +468,13 @@ public:
     \param indice les indices portes par le vecteur
     \param lowerStored indique le stockage de la matrice U ou L^T
   */
-  void solveUpperTriangularLeft(FullMatrix<T>* b, bool lowerStored=false) const;
+  void solveUpperTriangularLeft(FullMatrix<T>* b, bool unitriangular, bool lowerStored) const;
   /* Solve D x = b, in place with D a diagonal matrix.
 
      \param b Input: B, Output: X
    */
   void solveDiagonal(FullMatrix<T>* b) const;
+  void solveDiagonal(HMatrix<T>* b) const;
   /*! Resolution de This * x = b.
 
     \warning This doit etre factorisee avec \a HMatrix::luDecomposition() avant.
@@ -408,20 +484,43 @@ public:
 
     \warning This doit etre factorisee avec \a HMatrix::luDecomposition() avant.
    */
-  void solve(HMatrix<T>* b) const;
+  void solve(HMatrix<T>* b, hmat_factorization_t) const;
   /*! Resolution de This * x = b.
 
     \warning This doit etre factorisee avec \a HMatrix::ldltDecomposition() avant.
    */
   void solveLdlt(FullMatrix<T>* b) const ;
+  /*! Resolution de This * x = b.
+
+    \warning This doit etre factorisee avec \a HMatrix::lltDecomposition() avant.
+   */
+  void solveLlt(FullMatrix<T>* b) const ;
   /*! Triggers an assertion is the HMatrix contains any NaN.
    */
   void checkNan() const;
   /** Recursively set the isTriLower flag on this matrix */
   void setTriLower(bool value);
-public:
+
   const ClusterData* rows() const;
   const ClusterData* cols() const;
+
+  /*! \brief Return the number of children in the row dimension.
+    */
+  inline int nrChildRow() const {
+    return rows_->nrChild();
+  }
+
+  /*! \brief Return the number of children in the column dimension.
+    */
+  inline int nrChildCol() const {
+    return cols_->nrChild();
+  }
+  /*! \brief Destroy the HMatrix.
+    */
+  void destroy() {
+    delete this;
+  }
+
   /*! Return the child (i, j) of this.
 
     \warning do not use on a leaf !
@@ -430,11 +529,32 @@ public:
     \param j column
     \return the (i,j) child of this.
    */
-  inline HMatrix<T>* get(int i, int j) const {
-    return static_cast<HMatrix<T>*>(getChild(i + j * 2));
+  HMatrix<T>* get(int i, int j) const {
+    assert(i>=0 && i<nrChildRow());
+    assert(j>=0 && j<nrChildCol());
+    return this->getChild(i + j * nrChildRow());
   }
 
-public:
+  /*! Set the child (i, j) of this.
+
+    \warning do not use on a leaf !
+
+    \param i row index
+    \param j column index
+    \param child the child (i, j) of this.
+   */
+  using Tree<HMatrix<T> >::insertChild;
+  void insertChild(int i, int j, HMatrix<T>* child) {
+    insertChild(i+j*nrChildRow(), child) ;
+  }
+
+  void setClusterTrees(const ClusterTree* rows, const ClusterTree* cols);
+  HMatrix<T> * subset(const IndexSet * rows, const IndexSet * cols) const;
+
+  /* \brief Retrieve diagonal values.
+  */
+  void extractDiagonal(T* diag) const;
+
   /// Should try to coarsen the matrix at assembly
   static bool coarsening;
   /// Should recompress the matrix after assembly
@@ -447,35 +567,100 @@ public:
   static bool validationDump;
   /// Error threshold for the compression validation
   static double validationErrorThreshold;
-  HMatrixData<T> data;
-  bool isUpper, isLower;       /// symmetric, upper or lower stored
-  bool isTriUpper, isTriLower; /// upper/lower triangular
+  char isUpper:1, isLower:1,       /// symmetric, upper or lower stored
+       isTriUpper:1, isTriLower:1, /// upper/lower triangular
+       admissible:1, temporary:1,
+       ownClusterTree_:1;
+  LocalSettings localSettings;
 
-private:
-  /* \brief Resolution de X * D = B, avec D = this (matrice dont on ne tient compte que de la diagonale)
-     et B <- X
+  int rank() const {
+      assert(rank_ >= 0);
+      return rank_;
+  }
 
-     \param b la HMatrix en entree qui contiendra la solution en sortie
-  */
-  void getDiag(Vector<T>* diag, int start=0) const;
-#ifdef DEBUG_LDLT
-  /*  \brief verifie que la matrice est bien Lower i.e. avec des fils NULL au-dessus
+  RkMatrix<T> * rk() const {
+      assert(rank_ >= 0);
+      return rk_;
+  }
+
+  void rk(const FullMatrix<T> * a, const FullMatrix<T> * b, bool updateRank = true);
+
+  void rk(RkMatrix<T> * m) {
+      rk_ = m;
+      rank_ = m == NULL ? 0 : m->rank();
+  }
+
+  FullMatrix<T> * full() const {
+      assert(rank_ == FULL_BLOCK);
+      return full_;
+  }
+
+  void full(FullMatrix<T> * m) {
+      full_ = m;
+      rank_ = FULL_BLOCK;
+  }
+
+  bool isNull() const {
+      assert(rank_ >= FULL_BLOCK);
+      return rank_ == 0 || (rank_ == FULL_BLOCK && full_ == NULL);
+  }
+
+  bool isAssembled() const {
+      return rank_ > UNINITIALIZED_BLOCK;
+  }
+
+  /**
+   * Tag a not leaf block as assembled.
+   * Must only be called when all leaves of this block have been
+   * assembled (no coherency check).
    */
-  void assertLower();
-  /*  \brief verifie que la matrice est bien Upper i.e. avec des fils NULL au-dessous
+  void assembled() {
+      assert(!this->isLeaf());
+      rank_ = NONLEAF_BLOCK;
+  }
+
+  /**
+   * Tag an entire subtree (except the leaves) as assembled.
+   * (with recursion and coherency check: the leaves *must* already be tagged as assembled).
    */
-  void assertUpper();
-  /* \brief verifie que toutes les feuilles full de la hmatrice sont bien allouees
+  void assembledRecurse() {
+    if (!this->isLeaf()) {
+      for (int i=0 ; i<this->nrChild() ; i++)
+        if (this->getChild(i))
+          this->getChild(i)->assembledRecurse();
+      rank_ = NONLEAF_BLOCK;
+    }
+    assert(isAssembled());
+  }
+
+  const ClusterTree * rowsTree() const {
+      return rows_;
+  }
+
+  const ClusterTree * colsTree() const {
+      return cols_;
+  }
+
+  /**
+   * Convert this HMatrix to a string for debug.
+   * This is better than overriding << because it allows to use printf.
    */
-  void assertAllocFull();
-  /* \brief verifie juste que les matrices diagonales full ont bien une "diagonale" et sont isTriLower
-   */
-  bool assertLdlt() const;
-  /* \brief fait le produit LDL^T pour verifier la decomposition
-     Calcule la norme de L2 de deux matrices
-     \param originalCopy hmatrice avant decomposition LDL^T
-   */
-  void testLdlt(HMatrix<T> * originalCopy) ;
-#endif
+  std::string toString() const;
+
+  /*! \brief Return a short string describing the content of this HMatrix for debug (like: "HMatrix [320, 452]x[760, 890] norm=13.23442" or "uninitialized" if needed)
+    */
+  std::string description() const {
+    std::ostringstream convert;
+    convert << "HMatrix " << rows()->description() << "x" << cols()->description() ;
+    if (isAssembled())
+      convert << "norm=" << norm();
+    else
+      convert << "uninitialized";
+    return convert.str();
+  }
+
 };
+
+}  // end namespace hmat
+
 #endif
