@@ -25,9 +25,20 @@
 #define _LAPACK_OVERLOADS_HPP
 
 #include "data_types.hpp"
+#include <algorithm>
 
 #define F77_FUNC(a, b) a ##_
 
+namespace {
+inline int gesddRworkSize(char jobz, int m, int n) {
+    if (jobz == 'N')
+        return 7 * std::min(m, n);
+    else
+        return std::min(m, n) * std::max(5 * std::min(m, n) +
+            7, 2 * std::max(m, n) + 2 * std::min(m, n) + 1);
+}
+}
+
 namespace proxy_lapack {
 
 #ifndef DOXYGEN_SHOULD_SKIP_THIS
@@ -38,10 +49,10 @@ namespace proxy_lapack {
 #define _CGETRF_ F77_FUNC(cgetrf,CGETRF)
 #define _ZGETRF_ F77_FUNC(zgetrf,ZGETRF)
 extern "C" {
-void _SGETRF_(int*, int*, S_t*, int*, int*, int*);
-void _DGETRF_(int*, int*, D_t*, int*, int*, int*);
-void _CGETRF_(int*, int*, C_t*, int*, int*, int*);
-void _ZGETRF_(int*, int*, Z_t*, int*, int*, int*);
+void _SGETRF_(int*, int*, hmat::S_t*, int*, int*, int*);
+void _DGETRF_(int*, int*, hmat::D_t*, int*, int*, int*);
+void _CGETRF_(int*, int*, hmat::C_t*, int*, int*, int*);
+void _ZGETRF_(int*, int*, hmat::Z_t*, int*, int*, int*);
 }
 
 template <typename T>
@@ -49,28 +60,28 @@ int getrf(int m, int n, T* a, int lda, int* ipiv);
 
 template <>
 inline int
-getrf<S_t>(int m, int n, S_t* a, int lda, int* ipiv) {
+getrf<hmat::S_t>(int m, int n, hmat::S_t* a, int lda, int* ipiv) {
   int info = 0;
   _SGETRF_(&m, &n, a, &lda, ipiv, &info);
   return info;
 }
 template <>
 inline int
-getrf<D_t>(int m, int n, D_t* a, int lda, int* ipiv) {
+getrf<hmat::D_t>(int m, int n, hmat::D_t* a, int lda, int* ipiv) {
   int info = 0;
   _DGETRF_(&m, &n, a, &lda, ipiv, &info);
   return info;
 }
 template <>
 inline int
-getrf<C_t>(int m, int n, C_t* a, int lda, int* ipiv) {
+getrf<hmat::C_t>(int m, int n, hmat::C_t* a, int lda, int* ipiv) {
   int info = 0;
   _CGETRF_(&m, &n, a, &lda, ipiv, &info);
   return info;
 }
 template <>
 inline int
-getrf<Z_t>(int m, int n, Z_t* a, int lda, int* ipiv) {
+getrf<hmat::Z_t>(int m, int n, hmat::Z_t* a, int lda, int* ipiv) {
   int info = 0;
   _ZGETRF_(&m, &n, a, &lda, ipiv, &info);
   return info;
@@ -86,38 +97,38 @@ getrf<Z_t>(int m, int n, Z_t* a, int lda, int* ipiv) {
 #define _DGETRI_ F77_FUNC(dgetri,DGETRI)
 #define _CGETRI_ F77_FUNC(cgetri,CGETRI)
 #define _ZGETRI_ F77_FUNC(zgetri,ZGETRI)
-extern "C" void _SGETRI_(int*, S_t*, int*, const int*, S_t*, int*, int*);
-extern "C" void _DGETRI_(int*, D_t*, int*, const int*, D_t*, int*, int*);
-extern "C" void _CGETRI_(int*, C_t*, int*, const int*, C_t*, int*, int*);
-extern "C" void _ZGETRI_(int*, Z_t*, int*, const int*, Z_t*, int*, int*);
+extern "C" void _SGETRI_(int*, hmat::S_t*, int*, const int*, hmat::S_t*, int*, int*);
+extern "C" void _DGETRI_(int*, hmat::D_t*, int*, const int*, hmat::D_t*, int*, int*);
+extern "C" void _CGETRI_(int*, hmat::C_t*, int*, const int*, hmat::C_t*, int*, int*);
+extern "C" void _ZGETRI_(int*, hmat::Z_t*, int*, const int*, hmat::Z_t*, int*, int*);
 
 template<typename T>
 int getri(int n, T* a, int lda, const int* ipiv, T* work, int lwork);
 
 template<>
 inline int
-getri<S_t>(int n, S_t* a, int lda, const int* ipiv, S_t* work, int lwork) {
+getri<hmat::S_t>(int n, hmat::S_t* a, int lda, const int* ipiv, hmat::S_t* work, int lwork) {
   int info = 0;
   _SGETRI_(&n, a, &lda, ipiv, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-getri<D_t>(int n, D_t* a, int lda, const int* ipiv, D_t* work, int lwork) {
+getri<hmat::D_t>(int n, hmat::D_t* a, int lda, const int* ipiv, hmat::D_t* work, int lwork) {
   int info = 0;
   _DGETRI_(&n, a, &lda, ipiv, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-getri<C_t>(int n, C_t* a, int lda, const int* ipiv, C_t* work, int lwork) {
+getri<hmat::C_t>(int n, hmat::C_t* a, int lda, const int* ipiv, hmat::C_t* work, int lwork) {
   int info = 0;
   _CGETRI_(&n, a, &lda, ipiv, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-getri<Z_t>(int n, Z_t* a, int lda, const int* ipiv, Z_t* work, int lwork) {
+getri<hmat::Z_t>(int n, hmat::Z_t* a, int lda, const int* ipiv, hmat::Z_t* work, int lwork) {
   int info = 0;
   _ZGETRI_(&n, a, &lda, ipiv, work, &lwork, &info);
   return info;
@@ -134,38 +145,38 @@ getri<Z_t>(int n, Z_t* a, int lda, const int* ipiv, Z_t* work, int lwork) {
 #define _DGETRS_ F77_FUNC(dgetrs,DGETRS)
 #define _CGETRS_ F77_FUNC(cgetrs,CGETRS)
 #define _ZGETRS_ F77_FUNC(zgetrs,ZGETRS)
-extern "C" void _SGETRS_(char*, int*, int*, const S_t*, int*, const int*, S_t*, int*, int*);
-extern "C" void _DGETRS_(char*, int*, int*, const D_t*, int*, const int*, D_t*, int*, int*);
-extern "C" void _CGETRS_(char*, int*, int*, const C_t*, int*, const int*, C_t*, int*, int*);
-extern "C" void _ZGETRS_(char*, int*, int*, const Z_t*, int*, const int*, Z_t*, int*, int*);
+extern "C" void _SGETRS_(char*, int*, int*, const hmat::S_t*, int*, const int*, hmat::S_t*, int*, int*);
+extern "C" void _DGETRS_(char*, int*, int*, const hmat::D_t*, int*, const int*, hmat::D_t*, int*, int*);
+extern "C" void _CGETRS_(char*, int*, int*, const hmat::C_t*, int*, const int*, hmat::C_t*, int*, int*);
+extern "C" void _ZGETRS_(char*, int*, int*, const hmat::Z_t*, int*, const int*, hmat::Z_t*, int*, int*);
 
 template<typename T>
 int getrs(char trans, int n, int nrhs, const T* a, int lda, const int* ipiv, T* b, int ldb);
 
 template<>
 inline int
-getrs<S_t>(char trans, int n, int nrhs, const S_t* a, int lda, const int* ipiv, S_t* b, int ldb) {
+getrs<hmat::S_t>(char trans, int n, int nrhs, const hmat::S_t* a, int lda, const int* ipiv, hmat::S_t* b, int ldb) {
   int info = 0;
   _SGETRS_(&trans, &n, &nrhs, a, &lda, ipiv, b, &ldb, &info);
   return info;
 }
 template<>
 inline int
-getrs<D_t>(char trans, int n, int nrhs, const D_t* a, int lda, const int* ipiv, D_t* b, int ldb) {
+getrs<hmat::D_t>(char trans, int n, int nrhs, const hmat::D_t* a, int lda, const int* ipiv, hmat::D_t* b, int ldb) {
   int info = 0;
   _DGETRS_(&trans, &n, &nrhs, a, &lda, ipiv, b, &ldb, &info);
   return info;
 }
 template<>
 inline int
-getrs<C_t>(char trans, int n, int nrhs, const C_t* a, int lda, const int* ipiv, C_t* b, int ldb) {
+getrs<hmat::C_t>(char trans, int n, int nrhs, const hmat::C_t* a, int lda, const int* ipiv, hmat::C_t* b, int ldb) {
   int info = 0;
   _CGETRS_(&trans, &n, &nrhs, a, &lda, ipiv, b, &ldb, &info);
   return info;
 }
 template<>
 inline int
-getrs<Z_t>(char trans, int n, int nrhs, const Z_t* a, int lda, const int* ipiv, Z_t* b, int ldb) {
+getrs<hmat::Z_t>(char trans, int n, int nrhs, const hmat::Z_t* a, int lda, const int* ipiv, hmat::Z_t* b, int ldb) {
   int info = 0;
   _ZGETRS_(&trans, &n, &nrhs, a, &lda, ipiv, b, &ldb, &info);
   return info;
@@ -182,38 +193,38 @@ getrs<Z_t>(char trans, int n, int nrhs, const Z_t* a, int lda, const int* ipiv,
 #define _DGEQRF_ F77_FUNC(dgeqrf,DGEQRF)
 #define _CGEQRF_ F77_FUNC(cgeqrf,CGEQRF)
 #define _ZGEQRF_ F77_FUNC(zgeqrf,ZGEQRF)
-extern "C" void _SGEQRF_(int*, int*, S_t*, int*, S_t*, S_t*, int*, int*);
-extern "C" void _DGEQRF_(int*, int*, D_t*, int*, D_t*, D_t*, int*, int*);
-extern "C" void _CGEQRF_(int*, int*, C_t*, int*, C_t*, C_t*, int*, int*);
-extern "C" void _ZGEQRF_(int*, int*, Z_t*, int*, Z_t*, Z_t*, int*, int*);
+extern "C" void _SGEQRF_(int*, int*, hmat::S_t*, int*, hmat::S_t*, hmat::S_t*, int*, int*);
+extern "C" void _DGEQRF_(int*, int*, hmat::D_t*, int*, hmat::D_t*, hmat::D_t*, int*, int*);
+extern "C" void _CGEQRF_(int*, int*, hmat::C_t*, int*, hmat::C_t*, hmat::C_t*, int*, int*);
+extern "C" void _ZGEQRF_(int*, int*, hmat::Z_t*, int*, hmat::Z_t*, hmat::Z_t*, int*, int*);
 
 template<typename T>
 int geqrf(int m, int n, T* a, int lda, T* tau, T* work, int lwork);
 
 template<>
 inline int
-geqrf<S_t>(int m, int n, S_t* a, int lda, S_t* tau, S_t* work, int lwork) {
+geqrf<hmat::S_t>(int m, int n, hmat::S_t* a, int lda, hmat::S_t* tau, hmat::S_t* work, int lwork) {
   int info = 0;
   _SGEQRF_(&m, &n, a, &lda, tau, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-geqrf<D_t>(int m, int n, D_t* a, int lda, D_t* tau, D_t* work, int lwork) {
+geqrf<hmat::D_t>(int m, int n, hmat::D_t* a, int lda, hmat::D_t* tau, hmat::D_t* work, int lwork) {
   int info = 0;
   _DGEQRF_(&m, &n, a, &lda, tau, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-geqrf<C_t>(int m, int n, C_t* a, int lda, C_t* tau, C_t* work, int lwork) {
+geqrf<hmat::C_t>(int m, int n, hmat::C_t* a, int lda, hmat::C_t* tau, hmat::C_t* work, int lwork) {
   int info = 0;
   _CGEQRF_(&m, &n, a, &lda, tau, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-geqrf<Z_t>(int m, int n, Z_t* a, int lda, Z_t* tau, Z_t* work, int lwork) {
+geqrf<hmat::Z_t>(int m, int n, hmat::Z_t* a, int lda, hmat::Z_t* tau, hmat::Z_t* work, int lwork) {
   int info = 0;
   _ZGEQRF_(&m, &n, a, &lda, tau, work, &lwork, &info);
   return info;
@@ -223,6 +234,60 @@ geqrf<Z_t>(int m, int n, Z_t* a, int lda, Z_t* tau, Z_t* work, int lwork) {
 #undef _CGEQRF_
 #undef _ZGEQRF_
 
+/*      SUBROUTINE SGESDD( JOBZ, M, N, A, LDA, S, U, LDU, VT, LDVT,*/
+/*                         WORK, LWORK, IWORK, INFO )                     */
+/*      ZGESDD computes the singular value decomposition (SVD) of a complex */
+/*      M-by-N matrix A, optionally computing the left and/or right singular */
+/*      vectors   */
+#define _SGESDD_ F77_FUNC(sgesdd,SGESDD)
+#define _DGESDD_ F77_FUNC(dgesdd,DGESDD)
+#define _CGESDD_ F77_FUNC(cgesdd,CGESDD)
+#define _ZGESDD_ F77_FUNC(zgesdd,ZGESDD)
+extern "C" void _SGESDD_(char*, int*, int*, hmat::S_t*, int*,  float*, hmat::S_t*, int*, hmat::S_t*, int*, hmat::S_t*, int*, int*, int*);
+extern "C" void _DGESDD_(char*, int*, int*, hmat::D_t*, int*, double*, hmat::D_t*, int*, hmat::D_t*, int*, hmat::D_t*, int*, int*, int*);
+extern "C" void _CGESDD_(char*, int*, int*, hmat::C_t*, int*,  float*, hmat::C_t*, int*, hmat::C_t*, int*, hmat::C_t*, int*, float*, int*, int*);
+extern "C" void _ZGESDD_(char*, int*, int*, hmat::Z_t*, int*, double*, hmat::Z_t*, int*, hmat::Z_t*, int*, hmat::Z_t*, int*, double*, int*, int*);
+
+template<typename T, typename Treal>
+int gesdd(char jobz, int m, int n, T* a, int lda,  Treal* s, T* u, int ldu, T* vt, int ldvt, T* work, int lwork, int* iwork);
+
+template<>
+inline int
+gesdd<hmat::S_t, hmat::S_t>(char jobz, int m, int n, hmat::S_t* a, int lda,  hmat::S_t* s, hmat::S_t* u, int ldu, hmat::S_t* vt, int ldvt, hmat::S_t* work, int lwork, int* iwork) {
+  int info = 0;
+  _SGESDD_(&jobz, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, iwork, &info);
+  return info;
+}
+template<>
+inline int
+gesdd<hmat::D_t, hmat::D_t>(char jobz, int m, int n, hmat::D_t* a, int lda,  hmat::D_t* s, hmat::D_t* u, int ldu, hmat::D_t* vt, int ldvt, hmat::D_t* work, int lwork, int* iwork) {
+  int info = 0;
+  _DGESDD_(&jobz, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, iwork, &info);
+  return info;
+}
+template<>
+inline int
+gesdd<hmat::C_t, hmat::S_t>(char jobz, int m, int n, hmat::C_t* a, int lda,  hmat::S_t* s, hmat::C_t* u, int ldu, hmat::C_t* vt, int ldvt, hmat::C_t* work, int lwork, int* iwork) {
+  int info = 0;
+  hmat::S_t* rwork = (lwork == -1 ? NULL : new hmat::S_t[gesddRworkSize(jobz, m, n)]);
+  _CGESDD_(&jobz, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, rwork, iwork, &info);
+  if (rwork) delete [] rwork;
+  return info;
+}
+template<>
+inline int
+gesdd<hmat::Z_t, hmat::D_t>(char jobz, int m, int n, hmat::Z_t* a, int lda,  hmat::D_t* s, hmat::Z_t* u, int ldu, hmat::Z_t* vt, int ldvt, hmat::Z_t* work, int lwork, int* iwork) {
+  int info = 0;
+  hmat::D_t* rwork = (lwork == -1 ? NULL : new hmat::D_t[gesddRworkSize(jobz, m, n)]);
+  _ZGESDD_(&jobz, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, rwork, iwork, &info);
+  if (rwork) delete [] rwork;
+  return info;
+}
+#undef _SGESDD_
+#undef _DGESDD_
+#undef _CGESDD_
+#undef _ZGESDD_
+
 /*      SUBROUTINE SGESVD( JOBU, JOBVT, M, N, A, LDA, S, U, LDU, VT, LDVT,*/
 /*                         WORK, LWORK, INFO )                     */
 /*      ZGESVD computes the singular value decomposition (SVD) of a complex */
@@ -232,44 +297,44 @@ geqrf<Z_t>(int m, int n, Z_t* a, int lda, Z_t* tau, Z_t* work, int lwork) {
 #define _DGESVD_ F77_FUNC(dgesvd,DGESVD)
 #define _CGESVD_ F77_FUNC(cgesvd,CGESVD)
 #define _ZGESVD_ F77_FUNC(zgesvd,ZGESVD)
-extern "C" void _SGESVD_(char*, char*, int*, int*, S_t*, int*,  float*, S_t*, int*, S_t*, int*, S_t*, int*, int*);
-extern "C" void _DGESVD_(char*, char*, int*, int*, D_t*, int*, double*, D_t*, int*, D_t*, int*, D_t*, int*, int*);
-extern "C" void _CGESVD_(char*, char*, int*, int*, C_t*, int*,  float*, C_t*, int*, C_t*, int*, C_t*, int*,  float*, int*);
-extern "C" void _ZGESVD_(char*, char*, int*, int*, Z_t*, int*, double*, Z_t*, int*, Z_t*, int*, Z_t*, int*, double*, int*);
+extern "C" void _SGESVD_(char*, char*, int*, int*, hmat::S_t*, int*,  float*, hmat::S_t*, int*, hmat::S_t*, int*, hmat::S_t*, int*, int*);
+extern "C" void _DGESVD_(char*, char*, int*, int*, hmat::D_t*, int*, double*, hmat::D_t*, int*, hmat::D_t*, int*, hmat::D_t*, int*, int*);
+extern "C" void _CGESVD_(char*, char*, int*, int*, hmat::C_t*, int*,  float*, hmat::C_t*, int*, hmat::C_t*, int*, hmat::C_t*, int*,  float*, int*);
+extern "C" void _ZGESVD_(char*, char*, int*, int*, hmat::Z_t*, int*, double*, hmat::Z_t*, int*, hmat::Z_t*, int*, hmat::Z_t*, int*, double*, int*);
 
 template<typename T, typename Treal>
 int gesvd(char jobu, char jobvt, int m, int n, T* a, int lda,  Treal* s, T* u, int ldu, T* vt, int ldvt, T* work, int lwork);
 
 template<>
 inline int
-gesvd<S_t, S_t>(char jobu, char jobvt, int m, int n, S_t* a, int lda,  S_t* s, S_t* u, int ldu, S_t* vt, int ldvt, S_t* work, int lwork) {
+gesvd<hmat::S_t, hmat::S_t>(char jobu, char jobvt, int m, int n, hmat::S_t* a, int lda,  hmat::S_t* s, hmat::S_t* u, int ldu, hmat::S_t* vt, int ldvt, hmat::S_t* work, int lwork) {
   int info = 0;
   _SGESVD_(&jobu, &jobvt, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-gesvd<D_t, D_t>(char jobu, char jobvt, int m, int n, D_t* a, int lda,  D_t* s, D_t* u, int ldu, D_t* vt, int ldvt, D_t* work, int lwork) {
+gesvd<hmat::D_t, hmat::D_t>(char jobu, char jobvt, int m, int n, hmat::D_t* a, int lda,  hmat::D_t* s, hmat::D_t* u, int ldu, hmat::D_t* vt, int ldvt, hmat::D_t* work, int lwork) {
   int info = 0;
   _DGESVD_(&jobu, &jobvt, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-gesvd<C_t, S_t>(char jobu, char jobvt, int m, int n, C_t* a, int lda,  S_t* s, C_t* u, int ldu, C_t* vt, int ldvt, C_t* work, int lwork) {
+gesvd<hmat::C_t, hmat::S_t>(char jobu, char jobvt, int m, int n, hmat::C_t* a, int lda,  hmat::S_t* s, hmat::C_t* u, int ldu, hmat::C_t* vt, int ldvt, hmat::C_t* work, int lwork) {
   int info = 0;
   const int mn = (m < n ? m : n);
-  S_t* rwork = (lwork == -1 ? NULL : new S_t[5 * mn]);
+  hmat::S_t* rwork = (lwork == -1 ? NULL : new hmat::S_t[5 * mn]);
   _CGESVD_(&jobu, &jobvt, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, rwork, &info);
   if (rwork) delete [] rwork;
   return info;
 }
 template<>
 inline int
-gesvd<Z_t, D_t>(char jobu, char jobvt, int m, int n, Z_t* a, int lda,  D_t* s, Z_t* u, int ldu, Z_t* vt, int ldvt, Z_t* work, int lwork) {
+gesvd<hmat::Z_t, hmat::D_t>(char jobu, char jobvt, int m, int n, hmat::Z_t* a, int lda,  hmat::D_t* s, hmat::Z_t* u, int ldu, hmat::Z_t* vt, int ldvt, hmat::Z_t* work, int lwork) {
   int info = 0;
   const int mn = (m < n ? m : n);
-  D_t* rwork = (lwork == -1 ? NULL : new D_t[5 * mn]);
+  hmat::D_t* rwork = (lwork == -1 ? NULL : new hmat::D_t[5 * mn]);
   _ZGESVD_(&jobu, &jobvt, &m, &n, a, &lda, s, u, &ldu, vt, &ldvt, work, &lwork, rwork, &info);
   if (rwork) delete [] rwork;
   return info;
@@ -289,22 +354,22 @@ gesvd<Z_t, D_t>(char jobu, char jobvt, int m, int n, Z_t* a, int lda,  D_t* s, Z
 //   as returned by SGEQRF.
 #define _SORGQR_ F77_FUNC(sorgqr,SORGQR)
 #define _DORGQR_ F77_FUNC(dorgqr,DORGQR)
-extern "C" void _SORGQR_(int*, int*, int*, S_t*, int*, const S_t*, S_t*, int*, int*);
-extern "C" void _DORGQR_(int*, int*, int*, D_t*, int*, const D_t*, D_t*, int*, int*);
+extern "C" void _SORGQR_(int*, int*, int*, hmat::S_t*, int*, const hmat::S_t*, hmat::S_t*, int*, int*);
+extern "C" void _DORGQR_(int*, int*, int*, hmat::D_t*, int*, const hmat::D_t*, hmat::D_t*, int*, int*);
 
 template<typename T>
 int orgqr(int m, int n, int k, T* a, int lda, const T* tau, T* work, int lwork);
 
 template<>
 inline int
-orgqr<S_t>(int m, int n, int k, S_t* a, int lda, const S_t* tau, S_t* work, int lwork) {
+orgqr<hmat::S_t>(int m, int n, int k, hmat::S_t* a, int lda, const hmat::S_t* tau, hmat::S_t* work, int lwork) {
   int info = 0;
   _SORGQR_(&m, &n, &k, a, &lda, tau, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-orgqr<D_t>(int m, int n, int k, D_t* a, int lda, const D_t* tau, D_t* work, int lwork) {
+orgqr<hmat::D_t>(int m, int n, int k, hmat::D_t* a, int lda, const hmat::D_t* tau, hmat::D_t* work, int lwork) {
   int info = 0;
   _DORGQR_(&m, &n, &k, a, &lda, tau, work, &lwork, &info);
   return info;
@@ -322,22 +387,22 @@ orgqr<D_t>(int m, int n, int k, D_t* a, int lda, const D_t* tau, D_t* work, int
 //  as returned by CGEQRF.
 #define _CUNGQR_ F77_FUNC(cungqr,CUNGQR)
 #define _ZUNGQR_ F77_FUNC(zungqr,ZUNGQR)
-extern "C" void _CUNGQR_(int*, int*, int*, C_t*, int*, const C_t*, C_t*, int*, int*);
-extern "C" void _ZUNGQR_(int*, int*, int*, Z_t*, int*, const Z_t*, Z_t*, int*, int*);
+extern "C" void _CUNGQR_(int*, int*, int*, hmat::C_t*, int*, const hmat::C_t*, hmat::C_t*, int*, int*);
+extern "C" void _ZUNGQR_(int*, int*, int*, hmat::Z_t*, int*, const hmat::Z_t*, hmat::Z_t*, int*, int*);
 
 template<typename T>
 int ungqr(int m, int n, int k, T* a, int lda, const T* tau, T* work, int lwork);
 
 template<>
 inline int
-ungqr<C_t>(int m, int n, int k, C_t* a, int lda, const C_t* tau, C_t* work, int lwork) {
+ungqr<hmat::C_t>(int m, int n, int k, hmat::C_t* a, int lda, const hmat::C_t* tau, hmat::C_t* work, int lwork) {
   int info = 0;
   _CUNGQR_(&m, &n, &k, a, &lda, tau, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-ungqr<Z_t>(int m, int n, int k, Z_t* a, int lda, const Z_t* tau, Z_t* work, int lwork) {
+ungqr<hmat::Z_t>(int m, int n, int k, hmat::Z_t* a, int lda, const hmat::Z_t* tau, hmat::Z_t* work, int lwork) {
   int info = 0;
   _ZUNGQR_(&m, &n, &k, a, &lda, tau, work, &lwork, &info);
   return info;
@@ -362,22 +427,22 @@ ungqr<Z_t>(int m, int n, int k, Z_t* a, int lda, const Z_t* tau, Z_t* work, int
 //  if SIDE = 'R'.
 #define _SORMQR_ F77_FUNC(sormqr,SORMQR)
 #define _DORMQR_ F77_FUNC(dormqr,DORMQR)
-extern "C" void _SORMQR_(char*, char*, int*, int*, int*, const S_t*, int*, const S_t*, S_t*, int*, S_t*, int*, int*);
-extern "C" void _DORMQR_(char*, char*, int*, int*, int*, const D_t*, int*, const D_t*, D_t*, int*, D_t*, int*, int*);
+extern "C" void _SORMQR_(char*, char*, int*, int*, int*, const hmat::S_t*, int*, const hmat::S_t*, hmat::S_t*, int*, hmat::S_t*, int*, int*);
+extern "C" void _DORMQR_(char*, char*, int*, int*, int*, const hmat::D_t*, int*, const hmat::D_t*, hmat::D_t*, int*, hmat::D_t*, int*, int*);
 
 template<typename T>
 int ormqr(char side, char trans, int m, int n, int k, const T* a, int lda, const T* tau, T* c, int ldc, T* work, int lwork);
 
 template<>
 inline int
-ormqr<S_t>(char side, char trans, int m, int n, int k, const S_t* a, int lda, const S_t* tau, S_t* c, int ldc, S_t* work, int lwork) {
+ormqr<hmat::S_t>(char side, char trans, int m, int n, int k, const hmat::S_t* a, int lda, const hmat::S_t* tau, hmat::S_t* c, int ldc, hmat::S_t* work, int lwork) {
   int info = 0;
   _SORMQR_(&side, &trans, &m, &n, &k, a, &lda, tau, c, &ldc, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-ormqr<D_t>(char side, char trans, int m, int n, int k, const D_t* a, int lda, const D_t* tau, D_t* c, int ldc, D_t* work, int lwork) {
+ormqr<hmat::D_t>(char side, char trans, int m, int n, int k, const hmat::D_t* a, int lda, const hmat::D_t* tau, hmat::D_t* c, int ldc, hmat::D_t* work, int lwork) {
   int info = 0;
   _DORMQR_(&side, &trans, &m, &n, &k, a, &lda, tau, c, &ldc, work, &lwork, &info);
   return info;
@@ -402,22 +467,22 @@ ormqr<D_t>(char side, char trans, int m, int n, int k, const D_t* a, int lda, co
 //  if SIDE = 'R'.
 #define _CUNMQR_ F77_FUNC(cunmqr,CUNMQR)
 #define _ZUNMQR_ F77_FUNC(zunmqr,ZUNMQR)
-extern "C" void _CUNMQR_(char*, char*, int*, int*, int*, const C_t*, int*, const C_t*, C_t*, int*, C_t*, int*, int*);
-extern "C" void _ZUNMQR_(char*, char*, int*, int*, int*, const Z_t*, int*, const Z_t*, Z_t*, int*, Z_t*, int*, int*);
+extern "C" void _CUNMQR_(char*, char*, int*, int*, int*, const hmat::C_t*, int*, const hmat::C_t*, hmat::C_t*, int*, hmat::C_t*, int*, int*);
+extern "C" void _ZUNMQR_(char*, char*, int*, int*, int*, const hmat::Z_t*, int*, const hmat::Z_t*, hmat::Z_t*, int*, hmat::Z_t*, int*, int*);
 
 template<typename T>
 int unmqr(char side, char trans, int m, int n, int k, const T* a, int lda, const T* tau, T* c, int ldc, T* work, int lwork);
 
 template<>
 inline int
-unmqr<C_t>(char side, char trans, int m, int n, int k, const C_t* a, int lda, const C_t* tau, C_t* c, int ldc, C_t* work, int lwork) {
+unmqr<hmat::C_t>(char side, char trans, int m, int n, int k, const hmat::C_t* a, int lda, const hmat::C_t* tau, hmat::C_t* c, int ldc, hmat::C_t* work, int lwork) {
   int info = 0;
   _CUNMQR_(&side, &trans, &m, &n, &k, a, &lda, tau, c, &ldc, work, &lwork, &info);
   return info;
 }
 template<>
 inline int
-unmqr<Z_t>(char side, char trans, int m, int n, int k, const Z_t* a, int lda, const Z_t* tau, Z_t* c, int ldc, Z_t* work, int lwork) {
+unmqr<hmat::Z_t>(char side, char trans, int m, int n, int k, const hmat::Z_t* a, int lda, const hmat::Z_t* tau, hmat::Z_t* c, int ldc, hmat::Z_t* work, int lwork) {
   int info = 0;
   _ZUNMQR_(&side, &trans, &m, &n, &k, a, &lda, tau, c, &ldc, work, &lwork, &info);
   return info;
@@ -429,32 +494,32 @@ unmqr<Z_t>(char side, char trans, int m, int n, int k, const Z_t* a, int lda, co
 #define _DLASWP_ F77_FUNC(dlaswp,DLASWP)
 #define _CLASWP_ F77_FUNC(claswp,CLASWP)
 #define _ZLASWP_ F77_FUNC(zlaswp,ZLASWP)
-extern "C" void _SLASWP_(int*, S_t*, int*, int*, int*, const int*, int*);
-extern "C" void _DLASWP_(int*, D_t*, int*, int*, int*, const int*, int*);
-extern "C" void _CLASWP_(int*, C_t*, int*, int*, int*, const int*, int*);
-extern "C" void _ZLASWP_(int*, Z_t*, int*, int*, int*, const int*, int*);
+extern "C" void _SLASWP_(int*, hmat::S_t*, int*, int*, int*, const int*, int*);
+extern "C" void _DLASWP_(int*, hmat::D_t*, int*, int*, int*, const int*, int*);
+extern "C" void _CLASWP_(int*, hmat::C_t*, int*, int*, int*, const int*, int*);
+extern "C" void _ZLASWP_(int*, hmat::Z_t*, int*, int*, int*, const int*, int*);
 
 template<typename T>
 void laswp(int n, T* a, int lda, int k1, int k2, const int* ipiv, int incx);
 
 template<>
 inline void
-laswp<S_t>(int n, S_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
+laswp<hmat::S_t>(int n, hmat::S_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
   _SLASWP_(&n, a, &lda, &k1, &k2, ipiv, &incx);
 }
 template<>
 inline void
-laswp<D_t>(int n, D_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
+laswp<hmat::D_t>(int n, hmat::D_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
   _DLASWP_(&n, a, &lda, &k1, &k2, ipiv, &incx);
 }
 template<>
 inline void
-laswp<C_t>(int n, C_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
+laswp<hmat::C_t>(int n, hmat::C_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
   _CLASWP_(&n, a, &lda, &k1, &k2, ipiv, &incx);
 }
 template<>
 inline void
-laswp<Z_t>(int n, Z_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
+laswp<hmat::Z_t>(int n, hmat::Z_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
   _ZLASWP_(&n, a, &lda, &k1, &k2, ipiv, &incx);
 }
 #undef _SLASWP_
@@ -467,30 +532,30 @@ laswp<Z_t>(int n, Z_t* a, int lda, int k1, int k2, const int* ipiv, int incx) {
 #define _CPOTRF_ F77_FUNC(cpotrf,CPOTRF)
 #define _ZPOTRF_ F77_FUNC(zpotrf,ZPOTRF)
 extern "C" {
-void _SPOTRF_(char * uplo, int* n, S_t* a, int* lda, int* info);
-void _DPOTRF_(char * uplo, int* n, D_t* a, int* lda, int* info);
-void _CPOTRF_(char * uplo, int* n, C_t* a, int* lda, int* info);
-void _ZPOTRF_(char * uplo, int* n, Z_t* a, int* lda, int* info);
+void _SPOTRF_(char * uplo, int* n, hmat::S_t* a, int* lda, int* info);
+void _DPOTRF_(char * uplo, int* n, hmat::D_t* a, int* lda, int* info);
+void _CPOTRF_(char * uplo, int* n, hmat::C_t* a, int* lda, int* info);
+void _ZPOTRF_(char * uplo, int* n, hmat::Z_t* a, int* lda, int* info);
 }
 
 template <typename T> int potrf(char uplo, int n, T* a, int lda);
 
-template <> inline int potrf<S_t>(char uplo, int n, S_t* a, int lda) {
+template <> inline int potrf<hmat::S_t>(char uplo, int n, hmat::S_t* a, int lda) {
   int info = 0;
   _SPOTRF_(&uplo, &n, a, &lda, &info);
   return info;
 }
-template <> inline int potrf<D_t>(char uplo, int n, D_t* a, int lda) {
+template <> inline int potrf<hmat::D_t>(char uplo, int n, hmat::D_t* a, int lda) {
   int info = 0;
   _DPOTRF_(&uplo, &n, a, &lda, &info);
   return info;
 }
-template <> inline int potrf<C_t>(char uplo, int n, C_t* a, int lda) {
+template <> inline int potrf<hmat::C_t>(char uplo, int n, hmat::C_t* a, int lda) {
   int info = 0;
   _CPOTRF_(&uplo, &n, a, &lda, &info);
   return info;
 }
-template <> inline int potrf<Z_t>(char uplo, int n, Z_t* a, int lda) {
+template <> inline int potrf<hmat::Z_t>(char uplo, int n, hmat::Z_t* a, int lda) {
   int info = 0;
   _ZPOTRF_(&uplo, &n, a, &lda, &info);
   return info;
@@ -514,23 +579,23 @@ int or_un_mqr(char side, char trans, int m, int n, int k, const T* a, int lda, c
 
 template<>
 inline int
-or_un_mqr<S_t>(char side, char trans, int m, int n, int k, const S_t* a, int lda, const S_t* tau, S_t* c, int ldc, S_t* work, int lwork) {
+or_un_mqr<hmat::S_t>(char side, char trans, int m, int n, int k, const hmat::S_t* a, int lda, const hmat::S_t* tau, hmat::S_t* c, int ldc, hmat::S_t* work, int lwork) {
   return proxy_lapack::ormqr(side, trans, m, n, k, a, lda, tau, c, ldc, work, lwork);
 }
 template<>
 inline int
-or_un_mqr<D_t>(char side, char trans, int m, int n, int k, const D_t* a, int lda, const D_t* tau, D_t* c, int ldc, D_t* work, int lwork) {
+or_un_mqr<hmat::D_t>(char side, char trans, int m, int n, int k, const hmat::D_t* a, int lda, const hmat::D_t* tau, hmat::D_t* c, int ldc, hmat::D_t* work, int lwork) {
   return proxy_lapack::ormqr(side, trans, m, n, k, a, lda, tau, c, ldc, work, lwork);
 }
 template<>
 inline int
-or_un_mqr<C_t>(char side, char trans, int m, int n, int k, const C_t* a, int lda, const C_t* tau, C_t* c, int ldc, C_t* work, int lwork) {
+or_un_mqr<hmat::C_t>(char side, char trans, int m, int n, int k, const hmat::C_t* a, int lda, const hmat::C_t* tau, hmat::C_t* c, int ldc, hmat::C_t* work, int lwork) {
   const char t = (trans == 'N' ? 'N' : 'C');
   return proxy_lapack::unmqr(side, t, m, n, k, a, lda, tau, c, ldc, work, lwork);
 }
 template<>
 inline int
-or_un_mqr<Z_t>(char side, char trans, int m, int n, int k, const Z_t* a, int lda, const Z_t* tau, Z_t* c, int ldc, Z_t* work, int lwork) {
+or_un_mqr<hmat::Z_t>(char side, char trans, int m, int n, int k, const hmat::Z_t* a, int lda, const hmat::Z_t* tau, hmat::Z_t* c, int ldc, hmat::Z_t* work, int lwork) {
   const char t = (trans == 'N' ? 'N' : 'C');
   return proxy_lapack::unmqr(side, t, m, n, k, a, lda, tau, c, ldc, work, lwork);
 }
@@ -540,15 +605,25 @@ or_un_mqr<Z_t>(char side, char trans, int m, int n, int k, const Z_t* a, int lda
 namespace hmat {
 
 template<typename T> double real(const T x);
-template<> inline double real(const C_t x) {
+template<> inline double real(const hmat::C_t x) {
   return x.real();
 }
-template<> inline double real(const Z_t x) {
+template<> inline double real(const hmat::Z_t x) {
   return x.real();
 }
 template<typename T> inline double real(const T x) {
   return x;
 }
+template<typename T> double imag(const T x);
+template<> inline double imag(const hmat::C_t x) {
+  return x.imag();
+}
+template<> inline double imag(const hmat::Z_t x) {
+  return x.imag();
+}
+template<typename T> inline double imag(const T x) {
+  return 0;
+}
 
 }  // end namespace hmat
 
